from __future__ import annotations

import logging
import traceback
import warnings
from copy import deepcopy
from typing import TYPE_CHECKING, Callable

from great_expectations.compatibility.typing_extensions import override
from great_expectations.expectations.registry import get_renderer_impl
from great_expectations.render import (
    LegacyDiagnosticRendererType,
    RenderedTableContent,
)
from great_expectations.render.renderer.content_block.expectation_string import (
    ExpectationStringRenderer,
)

if TYPE_CHECKING:
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )

logger = logging.getLogger(__name__)


class ValidationResultsTableContentBlockRenderer(ExpectationStringRenderer):
    _content_block_type = "table"
    _rendered_component_type = RenderedTableContent
    _rendered_component_default_init_kwargs = {"table_options": {"search": True, "icon-size": "sm"}}

    _default_element_styling = {
        "default": {"classes": ["badge", "badge-secondary"]},
        "params": {"column": {"classes": ["badge", "badge-primary"]}},
    }

    _default_content_block_styling = {
        "body": {
            "classes": ["table"],
        },
        "classes": ["ml-2", "mr-2", "mt-0", "mb-0", "table-responsive"],
    }

    @classmethod
    def _get_custom_columns(cls, validation_results):
        custom_columns = []
        if (
            len(validation_results) > 0
            and "meta_properties_to_render" in validation_results[0].expectation_config.kwargs
            and validation_results[0].expectation_config.kwargs["meta_properties_to_render"]
            is not None
        ):
            custom_columns = list(
                validation_results[0].expectation_config.kwargs["meta_properties_to_render"].keys()
            )
        return sorted(custom_columns)

    @classmethod
    @override
    def _process_content_block(cls, content_block, has_failed_evr, render_object=None) -> None:
        super()._process_content_block(content_block, has_failed_evr)
        content_block.header_row = ["Status", "Expectation", "Observed Value"]
        content_block.header_row_options = {"Status": {"sortable": True}}

        # Add custom meta_properties_to_render header
        if render_object is not None:
            custom_columns = cls._get_custom_columns(render_object)
            content_block.header_row += custom_columns
            for column in custom_columns:
                content_block.header_row_options[column] = {"sortable": True}

        if has_failed_evr is False:
            styling = deepcopy(content_block.styling) if content_block.styling else {}
            if styling.get("classes"):
                styling["classes"].append("hide-succeeded-validations-column-section-target-child")
            else:
                styling["classes"] = ["hide-succeeded-validations-column-section-target-child"]

            content_block.styling = styling

    @override
    @classmethod
    def _get_content_block_fn(  # noqa: C901 # FIXME CoP
        cls,
        expectation_type: str,
        expectation_config: ExpectationConfiguration | None = None,
    ) -> Callable | None:
        content_block_fn = super()._get_content_block_fn(
            expectation_type=expectation_type, expectation_config=expectation_config
        )

        expectation_string_fn = content_block_fn
        if expectation_string_fn is None:
            expectation_string_fn = cls._get_legacy_v2_api_style_expectation_string_fn(
                expectation_type
            )
        if expectation_string_fn is None:
            expectation_string_fn = cls._missing_content_block_fn

        # This function wraps expect_* methods from ExpectationStringRenderer to generate table classes  # noqa: E501 # FIXME CoP
        def row_generator_fn(  # noqa: C901 # FIXME CoP
            configuration=None,
            result=None,
            runtime_configuration=None,
            **kwargs,
        ):
            eval_param_value_dict = kwargs.get("suite_parameters", None)
            # loading into suite parameters to be passed onto prescriptive renderer
            if eval_param_value_dict is not None:
                runtime_configuration["suite_parameters"] = eval_param_value_dict

            expectation = result.expectation_config
            expectation_string_cell = expectation_string_fn(
                configuration=expectation, runtime_configuration=runtime_configuration
            )

            status_icon_renderer = get_renderer_impl(
                object_name=expectation_type,
                renderer_type=LegacyDiagnosticRendererType.STATUS_ICON,
            )
            status_cell = (
                [status_icon_renderer[1](result=result)]
                if status_icon_renderer
                else [cls._diagnostic_status_icon_renderer(result=result)]
            )
            unexpected_statement = []
            unexpected_table = None
            observed_value = ["--"]

            data_docs_exception_message = """\
An unexpected Exception occurred during data docs rendering.  Because of this error, certain parts of data docs will \
not be rendered properly and/or may not appear altogether.  Please use the trace, included in this message, to \
diagnose and repair the underlying issue.  Detailed information follows:
            """  # noqa: E501 # FIXME CoP
            try:
                unexpected_statement_renderer = get_renderer_impl(
                    object_name=expectation_type,
                    renderer_type=LegacyDiagnosticRendererType.UNEXPECTED_STATEMENT,
                )
                unexpected_statement = (
                    unexpected_statement_renderer[1](result=result)
                    if unexpected_statement_renderer
                    else []
                )
            except Exception as e:
                exception_traceback = traceback.format_exc()
                exception_message = (
                    data_docs_exception_message
                    + f'{type(e).__name__}: "{e!s}".  Traceback: "{exception_traceback}".'
                )
                logger.error(exception_message)  # noqa: TRY400 # FIXME CoP
            try:
                unexpected_table_renderer = get_renderer_impl(
                    object_name=expectation_type,
                    renderer_type=LegacyDiagnosticRendererType.UNEXPECTED_TABLE,
                )
                unexpected_table = (
                    unexpected_table_renderer[1](result=result)
                    if unexpected_table_renderer
                    else None
                )
            except Exception as e:
                exception_traceback = traceback.format_exc()
                exception_message = (
                    data_docs_exception_message
                    + f'{type(e).__name__}: "{e!s}".  Traceback: "{exception_traceback}".'
                )
                logger.error(exception_message)  # noqa: TRY400 # FIXME CoP
            try:
                observed_value_renderer = get_renderer_impl(
                    object_name=expectation_type,
                    renderer_type=LegacyDiagnosticRendererType.OBSERVED_VALUE,
                )
                observed_value = [
                    observed_value_renderer[1](result=result)
                    if observed_value_renderer
                    else (
                        cls._get_legacy_v2_api_observed_value(expectation_string_fn, result) or "--"
                    )
                ]
            except Exception as e:
                exception_traceback = traceback.format_exc()
                exception_message = (
                    data_docs_exception_message
                    + f'{type(e).__name__}: "{e!s}".  Traceback: "{exception_traceback}".'
                )
                logger.error(exception_message)  # noqa: TRY400 # FIXME CoP

            # If the expectation has some unexpected values...:
            if unexpected_statement:
                expectation_string_cell += unexpected_statement
            if unexpected_table:
                expectation_string_cell += unexpected_table
            if len(expectation_string_cell) > 1:
                output_row = [status_cell + [expectation_string_cell] + observed_value]
            else:
                output_row = [status_cell + expectation_string_cell + observed_value]

            meta_properties_renderer = get_renderer_impl(
                object_name=expectation_type,
                renderer_type=LegacyDiagnosticRendererType.META_PROPERTIES,
            )
            if meta_properties_renderer:
                output_row[0] += meta_properties_renderer[1](result=result)

            return output_row

        return row_generator_fn

    @classmethod
    def _get_legacy_v2_api_style_expectation_string_fn(cls, expectation_type):
        legacy_expectation_string_fn = getattr(cls, expectation_type, None)
        if legacy_expectation_string_fn is None:
            # With the V2 API, expectation rendering was implemented by defining a method with the same name as the expectation.  # noqa: E501 # FIXME CoP
            # If no legacy rendering is present, return None.
            return None

        # deprecated-v0.13.28
        warnings.warn(
            "V2 API style custom rendering is deprecated as of v0.13.28 and is not fully supported anymore; "  # noqa: E501 # FIXME CoP
            "As it will be removed in v0.16, please transition to V3 API and associated rendering style",  # noqa: E501 # FIXME CoP
            DeprecationWarning,
        )

        def expectation_string_fn_with_legacy_translation(
            configuration: ExpectationConfiguration, runtime_configuration: dict
        ):
            if runtime_configuration is None:
                runtime_configuration = {}

            # With the V2 API, the expectation string function had a different signature; the below translates from the new signature to the legacy signature.  # noqa: E501 # FIXME CoP
            return legacy_expectation_string_fn(
                expectation=configuration,
                styling=runtime_configuration.get("styling", None),
                include_column_name=runtime_configuration.get("include_column_name", True),
            )

        return expectation_string_fn_with_legacy_translation

    @staticmethod
    def _get_legacy_v2_api_observed_value(expectation_string_fn, result):
        if expectation_string_fn.__name__ != "expectation_string_fn_with_legacy_translation":
            # If legacy V2 API style rendering is used, "expectation_string_fn" will be the method defined in the above "_get_legacy_v2_api_style_expectation_string_fn".  # noqa: E501 # FIXME CoP
            # If this isn't the case, return None, so we don't do any legacy logic.
            return None

        # With V2 API style rendering, the result had an "observed_value" entry that could be rendered.  # noqa: E501 # FIXME CoP
        return result["result"].get("observed_value")



"""
This module contains exceptions related to resource freshness.

A resource is considered "fresh" if it has been added to the DataContext and is up-to-date with
its persisted equivalent.

Please see the `FreshnessDiagnostics` class for how these exceptions are aggregated and raised.


Hierarchy:
    GreatExpectationsAggregateError
        ResourceFreshnessAggregateError
            CheckpointRelatedResourcesFreshnessError
            ValidationDefinitionRelatedResourcesFreshnessError

    ResourceFreshnessError
        ExpectationSuiteNotAddedError
        ExpectationSuiteNotFreshError
        BatchDefinitionNotAddedError
        BatchDefinitionNotFreshError
        ValidationDefinitionNotAddedError
        ValidationDefinitionNotFreshError
        CheckpointNotAddedError
        CheckpointNotFreshError
"""

from great_expectations.exceptions.exceptions import (
    DataContextError,
    GreatExpectationsAggregateError,
)


class ResourceFreshnessAggregateError(GreatExpectationsAggregateError):
    pass


class CheckpointRelatedResourcesFreshnessError(ResourceFreshnessAggregateError):
    pass


class ValidationDefinitionRelatedResourcesFreshnessError(ResourceFreshnessAggregateError):
    pass


class ResourceFreshnessError(DataContextError):
    pass


class ExpectationSuiteNotAddedError(ResourceFreshnessError):
    def __init__(self, name: str) -> None:
        super().__init__(
            f"ExpectationSuite '{name}' must be added to the DataContext before it can be updated. "
            "Please call `context.suites.add(<SUITE_OBJECT>)`, "
            "then try your action again."
        )


class ExpectationSuiteNotFreshError(ResourceFreshnessError):
    def __init__(self, name: str) -> None:
        super().__init__(
            f"ExpectationSuite '{name}' has changed since it has last been saved. "
            "Please update with `<SUITE_OBJECT>.save()`, then try your action again."
        )


class BatchDefinitionNotAddedError(ResourceFreshnessError):
    def __init__(self, name: str) -> None:
        super().__init__(
            f"BatchDefinition '{name}' must be added to the DataContext before it can be updated. "
            "Please update using the parent asset or data source, then try your action again."
        )


class BatchDefinitionNotFreshError(ResourceFreshnessError):
    def __init__(self, name: str) -> None:
        super().__init__(
            f"BatchDefinition '{name}' has changed since it has last been saved. "
            "Please update using the parent asset or data source, then try your action again."
        )


class ValidationDefinitionNotAddedError(ResourceFreshnessError):
    def __init__(self, name: str) -> None:
        super().__init__(
            f"ValidationDefinition '{name}' must be added to the DataContext before it can be updated. "  # noqa: E501 # FIXME CoP
            "Please call `context.validation_definitions.add(<VALIDATION_DEFINITION_OBJECT>)`, "
            "then try your action again."
        )


class ValidationDefinitionNotFreshError(ResourceFreshnessError):
    def __init__(self, name: str) -> None:
        super().__init__(
            f"ValidationDefinition '{name}' has changed since it has last been saved. "
            "Please update with `<VALIDATION_DEFINITION_OBJECT>.save()`, then try your action again."  # noqa: E501 # FIXME CoP
        )


class CheckpointNotAddedError(ResourceFreshnessError):
    def __init__(self, name: str) -> None:
        super().__init__(
            f"Checkpoint '{name}' must be added to the DataContext before it can be updated. "
            "Please call `context.checkpoints.add(<CHECKPOINT_OBJECT>)`, "
            "then try your action again."
        )


class CheckpointNotFreshError(ResourceFreshnessError):
    def __init__(self, name: str) -> None:
        super().__init__(
            f"Checkpoint '{name}' has changed since it has last been saved. "
            "Please update with `<CHECKPOINT_OBJECT>.save()`, then try your action again."
        )



import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    ALL_DATA_SOURCES,
    JUST_PANDAS_DATA_SOURCES,
)

NUM_COL = "all_numbers"
ALL_THE_SAME = "all_zeros"
MISSING_COL = "one_missing"

DATA = pd.DataFrame(
    {
        NUM_COL: [1, 1, 3],
        ALL_THE_SAME: [1, 1, 1],
        MISSING_COL: [None, -1, 1],
    },
    dtype="object",
)


@parameterize_batch_for_data_sources(data_source_configs=ALL_DATA_SOURCES, data=DATA)
def test_success_complete_results(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnStdevToBeBetween(column=NUM_COL, min_value=1.15, max_value=1.5)
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert result.success
    assert result.to_json_dict()["result"] == {"observed_value": pytest.approx(1.1547005383792517)}


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnStdevToBeBetween(column=ALL_THE_SAME, min_value=0, max_value=0),
            id="all_the_same",
        ),
        pytest.param(
            gxe.ExpectColumnStdevToBeBetween(column=MISSING_COL, min_value=1.414, max_value=1.415),
            id="missing_values",
        ),
        pytest.param(
            gxe.ExpectColumnStdevToBeBetween(column=NUM_COL, min_value=1),
            id="no_max",
        ),
        pytest.param(
            gxe.ExpectColumnStdevToBeBetween(column=NUM_COL, max_value=1.5),
            id="no_min",
        ),
        pytest.param(
            gxe.ExpectColumnStdevToBeBetween(
                column=NUM_COL, min_value=1, max_value=1.5, strict_min=True, strict_max=True
            ),
            id="strict_bounds",
        ),
        pytest.param(
            gxe.ExpectColumnStdevToBeBetween(column=NUM_COL),
            id="vacuous_truth",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success(
    batch_for_datasource: Batch, expectation: gxe.ExpectColumnStdevToBeBetween
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnStdevToBeBetween(column=ALL_THE_SAME, min_value=1, max_value=2),
            id="bad_range",
        ),
        pytest.param(
            gxe.ExpectColumnStdevToBeBetween(column=ALL_THE_SAME, min_value=0, strict_min=True),
            id="strict_min",
        ),
        pytest.param(
            gxe.ExpectColumnStdevToBeBetween(column=ALL_THE_SAME, max_value=0, strict_max=True),
            id="strict_max",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_failure(
    batch_for_datasource: Batch, expectation: gxe.ExpectColumnStdevToBeBetween
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert not result.success


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param(True, True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success_with_suite_param_strict_min_(
    batch_for_datasource: Batch, suite_param_value: bool, expected_result: bool
) -> None:
    suite_param_key = "test_expect_column_stdev_to_be_between"
    expectation = gxe.ExpectColumnStdevToBeBetween(
        column=NUM_COL,
        min_value=1,
        max_value=1.5,
        strict_min={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param(True, True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success_with_suite_param_strict_max_(
    batch_for_datasource: Batch, suite_param_value: bool, expected_result: bool
) -> None:
    suite_param_key = "test_expect_column_stdev_to_be_between"
    expectation = gxe.ExpectColumnStdevToBeBetween(
        column=NUM_COL,
        min_value=1,
        max_value=1.5,
        strict_max={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result



import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    ALL_DATA_SOURCES,
    JUST_PANDAS_DATA_SOURCES,
)

COL_NAME = "my_col"

ODD_COUNT = pd.DataFrame({COL_NAME: [1, 10, 11, None]}, dtype="object")
EVEN_COUNT = pd.DataFrame({COL_NAME: [1, 10, 11, 12, None]}, dtype="object")


@parameterize_batch_for_data_sources(data_source_configs=ALL_DATA_SOURCES, data=EVEN_COUNT)
def test_success_complete_results(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnMedianToBeBetween(column=COL_NAME, min_value=10, max_value=11)
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert result.success
    assert result.to_json_dict()["result"] == {"observed_value": 10.5}


@parameterize_batch_for_data_sources(data_source_configs=ALL_DATA_SOURCES, data=EVEN_COUNT)
def test_success_even_count(batch_for_datasource: Batch) -> None:
    """Ensure the median is calculated as the mean of the two middle values."""
    expectation = gxe.ExpectColumnMedianToBeBetween(column=COL_NAME, min_value=10.5, max_value=10.5)
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnMedianToBeBetween(column=COL_NAME),
            id="vacuous_success",
        ),
        pytest.param(
            gxe.ExpectColumnMedianToBeBetween(column=COL_NAME, min_value=9),
            id="just_min",
        ),
        pytest.param(
            gxe.ExpectColumnMedianToBeBetween(column=COL_NAME, max_value=11),
            id="just_max",
        ),
        pytest.param(
            gxe.ExpectColumnMedianToBeBetween(
                column=COL_NAME, min_value=9, max_value=11, strict_min=True, strict_max=True
            ),
            id="strict_min_and_max",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=ODD_COUNT)
def test_success(
    batch_for_datasource: Batch, expectation: gxe.ExpectColumnMedianToBeBetween
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnMedianToBeBetween(column=COL_NAME, min_value=11),
            id="just_min_fail",
        ),
        pytest.param(
            gxe.ExpectColumnMedianToBeBetween(
                column=COL_NAME, min_value=10, strict_min=True, max_value=11
            ),
            id="strict_min_fail",
        ),
        pytest.param(
            gxe.ExpectColumnMedianToBeBetween(column=COL_NAME, max_value=9),
            id="just_max_fail",
        ),
        pytest.param(
            gxe.ExpectColumnMedianToBeBetween(
                column=COL_NAME, min_value=9, max_value=10, strict_max=True
            ),
            id="strict_max_fail",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=ODD_COUNT)
def test_failure(
    batch_for_datasource: Batch, expectation: gxe.ExpectColumnMedianToBeBetween
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert not result.success


@parameterize_batch_for_data_sources(
    data_source_configs=JUST_PANDAS_DATA_SOURCES, data=pd.DataFrame({COL_NAME: []})
)
def test_no_data(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnMedianToBeBetween(
        column=COL_NAME, min_value=1, max_value=3, result_format=ResultFormat.SUMMARY
    )
    result = batch_for_datasource.validate(expectation)
    assert not result.success
    assert result.to_json_dict()["result"] == {"observed_value": None}


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param(True, True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=EVEN_COUNT)
def test_success_with_suite_param_strict_min_(
    batch_for_datasource: Batch, suite_param_value: bool, expected_result: bool
) -> None:
    suite_param_key = "expect_column_median_to_be_between"
    expectation = gxe.ExpectColumnMedianToBeBetween(
        column=COL_NAME,
        min_value=10,
        max_value=11,
        strict_min={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )

    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param(True, True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=EVEN_COUNT)
def test_success_with_suite_param_strict_max_(
    batch_for_datasource: Batch, suite_param_value: bool, expected_result: bool
) -> None:
    suite_param_key = "expect_column_median_to_be_between"
    expectation = gxe.ExpectColumnMedianToBeBetween(
        column=COL_NAME,
        min_value=10,
        max_value=11,
        strict_max={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result



from unittest.mock import Mock  # noqa: TID251 # FIXME CoP

import pytest

from great_expectations.data_context import AbstractDataContext
from great_expectations.data_context.data_context.context_factory import ProjectManager
from great_expectations.exceptions.exceptions import DataContextRequiredError


class TestProjectManagerStores:
    @pytest.mark.unit
    def test_get_expectations_store_success(self):
        context = Mock(spec=AbstractDataContext)
        project_manager = ProjectManager()
        project_manager.set_project(project=context)

        store = project_manager.get_expectations_store()

        assert store == context.expectations_store

    @pytest.mark.unit
    def test_get_expectations_store_fails_without_context(self):
        project_manager = ProjectManager()

        with pytest.raises(DataContextRequiredError):
            project_manager.get_expectations_store()

    @pytest.mark.unit
    def test_get_checkpoints_store_success(self):
        context = Mock(spec=AbstractDataContext)
        project_manager = ProjectManager()
        project_manager.set_project(project=context)

        store = project_manager.get_checkpoints_store()

        assert store == context.checkpoint_store

    @pytest.mark.unit
    def test_get_checkpoints_store_fails_without_context(self):
        project_manager = ProjectManager()

        with pytest.raises(DataContextRequiredError):
            project_manager.get_checkpoints_store()

    @pytest.mark.unit
    def test_get_validation_results_store_success(self):
        context = Mock(spec=AbstractDataContext)
        project_manager = ProjectManager()
        project_manager.set_project(project=context)

        store = project_manager.get_validation_results_store()

        assert store == context.validation_results_store

    @pytest.mark.unit
    def test_get_validation_results_store_fails_without_context(self):
        project_manager = ProjectManager()

        with pytest.raises(DataContextRequiredError):
            project_manager.get_validation_results_store()



from __future__ import annotations

import logging
from functools import wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Type,
    Union,
)

from great_expectations.compatibility.sqlalchemy import (
    sqlalchemy as sa,
)
from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.core.metric_function_types import (
    MetricPartialFunctionTypes,
)
from great_expectations.execution_engine import (
    ExecutionEngine,
    PandasExecutionEngine,
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.metrics.metric_provider import (
    metric_partial,
)
from great_expectations.expectations.metrics.util import (
    get_dbms_compatible_metric_domain_kwargs,
)

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from great_expectations.compatibility import sqlalchemy


def multicolumn_condition_partial(  # noqa: C901 #  16
    engine: Type[ExecutionEngine],
    partial_fn_type: Optional[MetricPartialFunctionTypes] = None,
    **kwargs,
):
    """Provides engine-specific support for authoring a metric_fn with a simplified signature. A
    multicolumn_condition_partial must provide a map function that evaluates to a boolean value; it will be used to
    provide supplemental metrics, such as the unexpected_value count, unexpected_values, and unexpected_rows.

    A metric function that is decorated as a multicolumn_condition_partial will be called with the engine-specific
    column_list type and any value_kwargs associated with the Metric for which the provider function is being declared.

    Args:
        engine:
        partial_fn_type:
        **kwargs:

    Returns:
        An annotated metric_function which will be called with a simplified signature.

    """  # noqa: E501 # FIXME CoP
    domain_type = MetricDomainTypes.MULTICOLUMN
    if issubclass(engine, PandasExecutionEngine):
        if partial_fn_type is None:
            partial_fn_type = MetricPartialFunctionTypes.MAP_CONDITION_SERIES

        partial_fn_type = MetricPartialFunctionTypes(partial_fn_type)
        if partial_fn_type not in [MetricPartialFunctionTypes.MAP_CONDITION_SERIES]:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                f"""PandasExecutionEngine only supports "{MetricPartialFunctionTypes.MAP_CONDITION_SERIES.value}" for \
"multicolumn_condition_partial" "partial_fn_type" property."""  # noqa: E501 # FIXME CoP
            )

        def wrapper(metric_fn: Callable):
            assert partial_fn_type is not None  # mypy has trouble type narrowing with closures

            @metric_partial(
                engine=engine,
                partial_fn_type=partial_fn_type,
                domain_type=domain_type,
                **kwargs,
            )
            @wraps(metric_fn)
            def inner_func(  # noqa: PLR0913 # FIXME CoP
                cls,
                execution_engine: PandasExecutionEngine,
                metric_domain_kwargs: dict,
                metric_value_kwargs: dict,
                metrics: Dict[str, Any],
                runtime_configuration: dict,
            ):
                metric_domain_kwargs = get_dbms_compatible_metric_domain_kwargs(
                    metric_domain_kwargs=metric_domain_kwargs,
                    batch_columns_list=metrics["table.columns"],
                )

                (
                    df,
                    compute_domain_kwargs,
                    accessor_domain_kwargs,
                ) = execution_engine.get_compute_domain(
                    domain_kwargs=metric_domain_kwargs, domain_type=domain_type
                )

                column_list: List[Union[str, sqlalchemy.quoted_name]] = accessor_domain_kwargs[
                    "column_list"
                ]

                meets_expectation_series = metric_fn(
                    cls,
                    df[column_list],
                    **metric_value_kwargs,
                    _metrics=metrics,
                )
                return (
                    ~meets_expectation_series,
                    compute_domain_kwargs,
                    accessor_domain_kwargs,
                )

            return inner_func

        return wrapper

    elif issubclass(engine, SqlAlchemyExecutionEngine):
        if partial_fn_type is None:
            partial_fn_type = MetricPartialFunctionTypes.MAP_CONDITION_FN

        partial_fn_type = MetricPartialFunctionTypes(partial_fn_type)
        if partial_fn_type not in [
            MetricPartialFunctionTypes.MAP_CONDITION_FN,
            MetricPartialFunctionTypes.WINDOW_CONDITION_FN,
        ]:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                f"""SqlAlchemyExecutionEngine only supports "{MetricPartialFunctionTypes.MAP_CONDITION_FN.value}" and \
"{MetricPartialFunctionTypes.WINDOW_CONDITION_FN.value}" for "multicolumn_condition_partial" "partial_fn_type" property.
"""  # noqa: E501 # FIXME CoP
            )

        def wrapper(metric_fn: Callable):
            assert partial_fn_type is not None  # mypy has trouble type narrowing with closures

            @metric_partial(
                engine=engine,
                partial_fn_type=partial_fn_type,
                domain_type=domain_type,
                **kwargs,
            )
            @wraps(metric_fn)
            def inner_func(  # noqa: PLR0913 # FIXME CoP
                cls,
                execution_engine: SqlAlchemyExecutionEngine,
                metric_domain_kwargs: dict,
                metric_value_kwargs: dict,
                metrics: Dict[str, Any],
                runtime_configuration: dict,
            ):
                metric_domain_kwargs = get_dbms_compatible_metric_domain_kwargs(
                    metric_domain_kwargs=metric_domain_kwargs,
                    batch_columns_list=metrics["table.columns"],
                )

                (
                    selectable,
                    compute_domain_kwargs,
                    accessor_domain_kwargs,
                ) = execution_engine.get_compute_domain(
                    domain_kwargs=metric_domain_kwargs, domain_type=domain_type
                )

                column_list: List[Union[str, sqlalchemy.quoted_name]] = accessor_domain_kwargs[
                    "column_list"
                ]

                sqlalchemy_engine: sqlalchemy.Engine = execution_engine.engine

                column_selector = [sa.column(column_name) for column_name in column_list]  # type: ignore[var-annotated] # FIXME CoP
                dialect = execution_engine.dialect_module
                expected_condition = metric_fn(
                    cls,
                    column_selector,
                    **metric_value_kwargs,
                    _dialect=dialect,
                    _table=selectable,
                    _sqlalchemy_engine=sqlalchemy_engine,
                    _metrics=metrics,
                )

                unexpected_condition = sa.not_(expected_condition)
                return (
                    unexpected_condition,
                    compute_domain_kwargs,
                    accessor_domain_kwargs,
                )

            return inner_func

        return wrapper

    elif issubclass(engine, SparkDFExecutionEngine):
        if partial_fn_type is None:
            partial_fn_type = MetricPartialFunctionTypes.MAP_CONDITION_FN

        partial_fn_type = MetricPartialFunctionTypes(partial_fn_type)
        if partial_fn_type not in [
            MetricPartialFunctionTypes.MAP_CONDITION_FN,
            MetricPartialFunctionTypes.WINDOW_CONDITION_FN,
        ]:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                f"""SparkDFExecutionEngine only supports "{MetricPartialFunctionTypes.MAP_CONDITION_FN.value}" and \
"{MetricPartialFunctionTypes.WINDOW_CONDITION_FN.value}" for "multicolumn_condition_partial" "partial_fn_type" property.
"""  # noqa: E501 # FIXME CoP
            )

        def wrapper(metric_fn: Callable):
            @metric_partial(
                engine=engine,
                partial_fn_type=partial_fn_type,
                domain_type=domain_type,
                **kwargs,
            )
            @wraps(metric_fn)
            def inner_func(  # noqa: PLR0913 # FIXME CoP
                cls,
                execution_engine: SparkDFExecutionEngine,
                metric_domain_kwargs: dict,
                metric_value_kwargs: dict,
                metrics: Dict[str, Any],
                runtime_configuration: dict,
            ):
                metric_domain_kwargs = get_dbms_compatible_metric_domain_kwargs(
                    metric_domain_kwargs=metric_domain_kwargs,
                    batch_columns_list=metrics["table.columns"],
                )

                (
                    data,
                    compute_domain_kwargs,
                    accessor_domain_kwargs,
                ) = execution_engine.get_compute_domain(
                    domain_kwargs=metric_domain_kwargs, domain_type=domain_type
                )

                column_list: List[Union[str, sqlalchemy.quoted_name]] = accessor_domain_kwargs[
                    "column_list"
                ]

                expected_condition = metric_fn(
                    cls,
                    data[column_list],
                    **metric_value_kwargs,
                    _metrics=metrics,
                )
                return (
                    ~expected_condition,
                    compute_domain_kwargs,
                    accessor_domain_kwargs,
                )

            return inner_func

        return wrapper

    else:
        raise ValueError(  # noqa: TRY003, TRY004 # FIXME CoP
            'Unsupported engine for "multicolumn_condition_partial" metric function decorator.'
        )



# isort:skip_file

# Set up version information immediately
from ._version import get_versions as _get_versions

__version__ = _get_versions()["version"]
del _get_versions

# Submodules must be imported first to avoid circular dependencies
# with data_context being the first
from . import data_context  # isort:skip
from . import core
from . import exceptions
from . import expectations
from . import checkpoint

# Top-level functions/classes promoted to the gx namespace
from great_expectations.data_context.data_context.context_factory import get_context
from great_expectations.checkpoint import Checkpoint
from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.core.result_format import ResultFormat
from great_expectations.core.run_identifier import RunIdentifier
from great_expectations.core.validation_definition import ValidationDefinition

# # By placing this registry function in our top-level __init__,  we ensure that all
# # GX workflows have populated expectation registries before they are used.
from great_expectations.expectations.registry import (
    register_core_expectations as _register_core_expectations,
    register_core_metrics as _register_core_metrics,
)

_register_core_metrics()
_register_core_expectations()

del _register_core_metrics
del _register_core_expectations



from __future__ import annotations

from typing import Final

from great_expectations.compatibility.not_imported import NotImported

SNOWFLAKE_NOT_IMPORTED = NotImported(
    "snowflake connection components are not installed, please 'pip install snowflake-sqlalchemy snowflake-connector-python'"  # noqa: E501 # FIXME CoP
)

try:
    import snowflake
except ImportError:
    snowflake = SNOWFLAKE_NOT_IMPORTED

try:
    from snowflake.sqlalchemy import URL
except ImportError:
    URL = SNOWFLAKE_NOT_IMPORTED

try:
    import snowflake.sqlalchemy as snowflakesqlalchemy
except (ImportError, AttributeError):
    snowflakesqlalchemy = SNOWFLAKE_NOT_IMPORTED

try:
    import snowflake.sqlalchemy.snowdialect as snowflakedialect
except (ImportError, AttributeError):
    snowflakedialect = SNOWFLAKE_NOT_IMPORTED

try:
    import snowflake.sqlalchemy.custom_types as snowflaketypes
except (ImportError, AttributeError):
    snowflaketypes = SNOWFLAKE_NOT_IMPORTED

IS_SNOWFLAKE_INSTALLED: Final[bool] = snowflake is not SNOWFLAKE_NOT_IMPORTED

try:
    from snowflake.sqlalchemy.custom_types import ARRAY
except (ImportError, AttributeError):
    ARRAY = SNOWFLAKE_NOT_IMPORTED

try:
    from snowflake.sqlalchemy.custom_types import BYTEINT
except (ImportError, AttributeError):
    BYTEINT = SNOWFLAKE_NOT_IMPORTED


try:
    from snowflake.sqlalchemy.custom_types import CHARACTER
except (ImportError, AttributeError):
    CHARACTER = SNOWFLAKE_NOT_IMPORTED


try:
    from snowflake.sqlalchemy.custom_types import DEC
except (ImportError, AttributeError):
    DEC = SNOWFLAKE_NOT_IMPORTED


try:
    from snowflake.sqlalchemy.custom_types import FIXED
except (ImportError, AttributeError):
    FIXED = SNOWFLAKE_NOT_IMPORTED

try:
    from snowflake.sqlalchemy.custom_types import GEOGRAPHY
except (ImportError, AttributeError):
    GEOGRAPHY = SNOWFLAKE_NOT_IMPORTED


try:
    from snowflake.sqlalchemy.custom_types import GEOMETRY
except (ImportError, AttributeError):
    GEOMETRY = SNOWFLAKE_NOT_IMPORTED


try:
    from snowflake.sqlalchemy.custom_types import NUMBER
except (ImportError, AttributeError):
    NUMBER = SNOWFLAKE_NOT_IMPORTED


try:
    from snowflake.sqlalchemy.custom_types import OBJECT
except (ImportError, AttributeError):
    OBJECT = SNOWFLAKE_NOT_IMPORTED


try:
    from snowflake.sqlalchemy.custom_types import STRING
except (ImportError, AttributeError):
    STRING = SNOWFLAKE_NOT_IMPORTED


try:
    from snowflake.sqlalchemy.custom_types import TEXT
except (ImportError, AttributeError):
    TEXT = SNOWFLAKE_NOT_IMPORTED


try:
    from snowflake.sqlalchemy.custom_types import TIMESTAMP_LTZ
except (ImportError, AttributeError):
    TIMESTAMP_LTZ = SNOWFLAKE_NOT_IMPORTED


try:
    from snowflake.sqlalchemy.custom_types import TIMESTAMP_NTZ
except (ImportError, AttributeError):
    TIMESTAMP_NTZ = SNOWFLAKE_NOT_IMPORTED


try:
    from snowflake.sqlalchemy.custom_types import TIMESTAMP_TZ
except (ImportError, AttributeError):
    TIMESTAMP_TZ = SNOWFLAKE_NOT_IMPORTED


try:
    from snowflake.sqlalchemy.custom_types import TINYINT
except (ImportError, AttributeError):
    TINYINT = SNOWFLAKE_NOT_IMPORTED


try:
    from snowflake.sqlalchemy.custom_types import VARBINARY
except (ImportError, AttributeError):
    VARBINARY = SNOWFLAKE_NOT_IMPORTED


try:
    from snowflake.sqlalchemy.custom_types import VARIANT
except (ImportError, AttributeError):
    VARIANT = SNOWFLAKE_NOT_IMPORTED

try:
    from snowflake.sqlalchemy.custom_types import DOUBLE
except (ImportError, AttributeError):
    DOUBLE = SNOWFLAKE_NOT_IMPORTED

try:
    from snowflake.sqlalchemy.custom_types import SnowflakeType
except (ImportError, AttributeError):
    SnowflakeType = SNOWFLAKE_NOT_IMPORTED


class SNOWFLAKE_TYPES:
    """Namespace for Snowflake dialect types."""

    ARRAY = ARRAY
    BYTEINT = BYTEINT
    CHARACTER = CHARACTER
    DEC = DEC
    DOUBLE = DOUBLE
    FIXED = FIXED
    GEOGRAPHY = GEOGRAPHY
    GEOMETRY = GEOMETRY
    NUMBER = NUMBER
    OBJECT = OBJECT
    STRING = STRING
    TEXT = TEXT
    TIMESTAMP_LTZ = TIMESTAMP_LTZ
    TIMESTAMP_NTZ = TIMESTAMP_NTZ
    TIMESTAMP_TZ = TIMESTAMP_TZ
    TINYINT = TINYINT
    VARBINARY = VARBINARY
    VARIANT = VARIANT



from __future__ import annotations

import inspect
import random
from pprint import pformat as pf
from typing import Any, Callable, Final, Literal, Protocol

import pytest

from great_expectations.compatibility import pydantic
from great_expectations.datasource.fluent import (
    DataAsset,
    Datasource,
    GxInvalidDatasourceWarning,
    InvalidAsset,
    InvalidDatasource,
    TestConnectionError,
)
from great_expectations.datasource.fluent.sources import DataSourceManager

pytestmark = pytest.mark.unit

_EXCLUDE_METHODS: Final[set[str]] = {
    # we don't care about overriding these methods for InvalidDatasource
    "copy",
    "delete_asset",
    "delete_batch_definition",
    "dict",
    "get_asset_names",
    "get_assets_as_dict",
    "get_batch_definition",  # DataAsset
    "get_execution_engine",
    "json",
    "update_batch_definition_field_set",  # DataAsset
    "yaml",
}
DATASOURCE_PUBLIC_METHODS: Final[list[str]] = [
    f[0]
    for f in inspect.getmembers(Datasource, predicate=inspect.isfunction)
    if not f[0].startswith("_") and f[0] not in _EXCLUDE_METHODS
]
DATA_ASSET_PUBLIC_METHODS: Final[list[str]] = [
    f[0]
    for f in inspect.getmembers(DataAsset, predicate=inspect.isfunction)
    if not f[0].startswith("_") and f[0] not in _EXCLUDE_METHODS
]


class TestPublicMethodsAreOverridden:
    """
    Ensure that applicable Datasource/DataAsset public methods are overridden.
    Applicable public methods are those that would typically be called when a users is trying to run
    some action on a Datasource and would want to know if the Datasource is invalid.

    If a method is not overridden, it will be inherited from the base class and will not be present
    in the InvalidDatasource.__dict__.
    """

    @pytest.mark.parametrize("base_ds_method_name", DATASOURCE_PUBLIC_METHODS)
    def test_datasource_methods(self, base_ds_method_name: str):
        """Ensure that InvalidDatasource overrides the applicable Datasource methods."""
        method = getattr(InvalidDatasource, base_ds_method_name, None)
        assert method, f"Expected {base_ds_method_name} to be defined on InvalidDatasource"
        with pytest.raises(TypeError):
            method()

    @pytest.mark.parametrize("base_ds_method_name", DATA_ASSET_PUBLIC_METHODS)
    def test_data_asset(self, base_ds_method_name: str):
        """Ensure that InvalidAsset overrides the applicable DataAsset methods."""
        method = getattr(InvalidAsset, base_ds_method_name, None)
        assert method, f"Expected {base_ds_method_name} to be defined on InvalidAsset"
        with pytest.raises(TypeError):
            method()


@pytest.fixture(scope="module")
def datasource_fields() -> set[str]:
    """Return a set of all the fields in the base Datasource model."""
    return set(Datasource.__fields__.keys())


@pytest.fixture(scope="module")
def data_asset_fields() -> set[str]:
    """Return a set of all the fields in the base DataAsset model."""
    return set(DataAsset.__fields__.keys())


class InvalidDSFactory(Protocol):
    """
    Accept a datasource config and return an InvalidDatasource instance.
    Raises an error if the config was valid.
    """

    def __call__(
        self, config: dict[Literal["name", "type", "assets"] | Any, Any]
    ) -> InvalidDatasource: ...


@pytest.fixture
def invalid_datasource_factory() -> InvalidDSFactory:
    def _invalid_ds_fct(config: dict) -> InvalidDatasource:
        try:
            ds_type: type[Datasource] = DataSourceManager.type_lookup[config["type"]]
            ds_type(**config)
        except (pydantic.ValidationError, LookupError) as config_error:
            return InvalidDatasource(**config, config_error=config_error)
        raise ValueError("The Datasource was valid")

    return _invalid_ds_fct


@pytest.mark.parametrize(
    "invalid_ds_cfg",
    [
        pytest.param(
            {
                "name": "my pg",
                "type": "postgres",
                "connection_string": "postmalone+psycopg2://postgres:@localhost/test_database",
            },
            id="invalid conn str",
        ),
        pytest.param(
            {
                "name": "my pg + asset",
                "type": "postgres",
                "connection_string": "postgresql+psycopg2://postgres:@localhost/test_database",
                "assets": [
                    {
                        "name": "my_bad_asset",
                        "type": "table",
                        "query": "table assets don't have a query",
                    }
                ],
            },
            id="invalid asset",
        ),
        pytest.param(
            {
                "name": "my snowflake",
                "type": "snowflake",
                "connection_string": "${MY_CONN_STR}",
                "user": "invalid_extra_field",
                "assets": [{"name": "my_asset", "type": "table", "table_name": "foobar"}],
            },
            id="extra field",
        ),
        pytest.param(
            {
                "name": "my pandas",
                "type": "pandas_filesystem",
                "assets": [
                    {"name": "my_asset", "type": "csv"},
                    {"name": "invalid_asset_type", "type": "whoops"},
                ],
            },
            id="pandas asset lookup error",
        ),
        pytest.param(
            {
                "name": "who knows",
                "type": "whoops",
            },
            id="datasource type lookup error",
        ),
    ],
)
class TestInvalidDatasource:
    def test_connection_raises_informative_error(
        self,
        invalid_ds_cfg: dict,
        invalid_datasource_factory: InvalidDSFactory,
    ):
        invalid_ds = invalid_datasource_factory(invalid_ds_cfg)
        print(invalid_ds)

        with pytest.raises(TestConnectionError) as conn_err:
            invalid_ds.test_connection()
        print(f"{conn_err.value!r}\n >-- caused by -->\n{conn_err.value.__cause__!r}")
        assert invalid_ds.config_error == conn_err.value.__cause__

    def test_get_batch_list_raises_informative_error(
        self,
        invalid_ds_cfg: dict,
        invalid_datasource_factory: Callable[
            [dict[Literal["name", "type", "assets"] | Any, Any]], InvalidDatasource
        ],
    ):
        invalid_ds = invalid_datasource_factory(invalid_ds_cfg)
        with pytest.raises(TypeError) as err:
            invalid_ds.get_batch({})  # type: ignore[arg-type] # expect error
        assert invalid_ds.config_error == err.value.__cause__

    def test_random_attribute_access_raises_informative_error(
        self, invalid_ds_cfg: dict, invalid_datasource_factory: InvalidDSFactory
    ):
        invalid_ds = invalid_datasource_factory(invalid_ds_cfg)
        with pytest.raises(TypeError) as err:
            _ = invalid_ds.random_attribute
        assert invalid_ds.config_error == err.value.__cause__

    @pytest.mark.parametrize("attr_name", ["name", "id", "type", "assets"])
    def test_base_datasource_attribute_does_not_error(
        self,
        invalid_ds_cfg: dict,
        invalid_datasource_factory: InvalidDSFactory,
        attr_name: str,
    ):
        invalid_ds = invalid_datasource_factory(invalid_ds_cfg)
        attr_value = getattr(invalid_ds, attr_name)
        print(attr_name, attr_value)

    def test_get_asset_raises_warning(
        self,
        invalid_ds_cfg: dict,
        invalid_datasource_factory: Callable[
            [dict[Literal["name", "type", "assets"] | Any, Any]], InvalidDatasource
        ],
    ):
        invalid_ds = invalid_datasource_factory(invalid_ds_cfg)
        for asset in invalid_ds.assets:
            with pytest.warns(GxInvalidDatasourceWarning):
                invalid_asset = invalid_ds.get_asset(asset.name)
                assert invalid_asset, "No asset was returned"

            with pytest.raises(TestConnectionError):
                invalid_asset.test_connection()

    def test_extra_fields_are_ignored(
        self,
        datasource_fields: set[str],
        data_asset_fields: set[str],
        invalid_ds_cfg: dict,
        invalid_datasource_factory: Callable[
            [dict[Literal["name", "type", "assets"] | Any, Any]], InvalidDatasource
        ],
    ):
        """
        Ensure that extra fields are ignored when creating the InvalidDatasource instance.
        These fields could include secrets or other sensitive information that should not be included
        in the InvalidDatasource instance.

        Standard fields such as `type`, `name`, `id` etc. should be included in the InvalidDatasource instance and should
        never be sensitive.
        """  # noqa: E501 # FIXME CoP
        print(f"Datasource config:\n{pf(invalid_ds_cfg)}")
        invalid_ds = invalid_datasource_factory(invalid_ds_cfg)

        ds_dict = invalid_ds._json_dict()
        print(f"\nInvalidDatasource dict:\n{pf(ds_dict)}")

        assert set(ds_dict.keys()) >= {
            "name",
            "type",
        }, "Expected standard fields to be present"

        extra_ds_fields = set(invalid_ds_cfg.keys()) - datasource_fields
        for field in extra_ds_fields:
            assert field not in ds_dict, f"Expected `{field}` to be ignored"

        for asset in invalid_ds.assets:
            asset_dict = asset.dict()
            extra_asset_fields = set(asset_dict.keys()) - data_asset_fields
            for field in extra_asset_fields:
                assert field not in asset_dict, f"Expected asset `{field}` to be ignored"


@pytest.fixture
def rand_invalid_datasource_with_assets(
    invalid_datasource_factory: InvalidDSFactory,
) -> InvalidDatasource:
    random_ds_type = random.choice([t for t in DataSourceManager.type_lookup.type_names()])
    invalid_ds = invalid_datasource_factory(
        {
            "name": "my invalid ds",
            "type": random_ds_type,
            "connection_string": "postgresql+psycopg2://postgres:@localhost/test_database",
            "assets": [
                {"name": "definitely_invalid", "type": "NOT_A_VALID_TYPE"},
                {"name": "maybe_valid", "type": "table", "table_name": "my_table"},
                {"name": "maybe_valid_2", "type": "csv", "sep": "|"},
                {"name": "missing type"},
            ],
        }
    )
    return invalid_ds


class TestInvalidDataAsset:
    def test_connection_raises_informative_error(
        self, invalid_datasource_factory: InvalidDSFactory
    ):
        random_ds_type = random.choice([t for t in DataSourceManager.type_lookup.type_names()])
        print(f"{random_ds_type=}")
        invalid_datasource: InvalidDatasource = invalid_datasource_factory(
            {
                "name": "my invalid ds",
                "type": random_ds_type,
                "foo": "bar",  # regardless of the type this extra field should make the datasource invalid  # noqa: E501 # FIXME CoP
                "assets": [
                    {"name": "definitely_invalid", "type": "NOT_A_VALID_TYPE"},
                    {"name": "maybe_valid", "type": "table", "table_name": "my_table"},
                    {"name": "maybe_valid_2", "type": "csv", "sep": "|"},
                    {"name": "missing type"},
                ],
            }
        )
        print(invalid_datasource)

        assert invalid_datasource.assets, "Expected assets to be present"
        for invalid_asset in invalid_datasource.assets:
            with pytest.raises(TestConnectionError) as conn_err:
                invalid_asset.test_connection()
                assert invalid_datasource.config_error == conn_err.value.__cause__

    @pytest.mark.parametrize("attr_name", ["name", "id", "type"])
    def test_base_data_asset_attribute_does_not_error(
        self, rand_invalid_datasource_with_assets: InvalidDatasource, attr_name: str
    ):
        assert rand_invalid_datasource_with_assets.assets, "Expected assets to be present"
        for asset in rand_invalid_datasource_with_assets.assets:
            value = getattr(asset, attr_name)
            print(attr_name, value)


if __name__ == "__main__":
    pytest.main(["-vv", __file__])



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Optional, Type, Union

import altair as alt
import numpy as np
import pandas as pd
from scipy import stats

from great_expectations.compatibility import pydantic
from great_expectations.core.suite_parameters import (
    SuiteParameterDict,  # noqa: TC001 # FIXME CoP
)
from great_expectations.core.types import Comparable  # noqa: TC001 # FIXME CoP
from great_expectations.execution_engine.util import (
    is_valid_categorical_partition_object,
    is_valid_partition_object,
)
from great_expectations.expectations.expectation import (
    COLUMN_DESCRIPTION,
    ColumnAggregateExpectation,
    render_suite_parameter_string,
)
from great_expectations.expectations.metadata_types import DataQualityIssues, SupportedDataSources
from great_expectations.render import (
    AtomicDiagnosticRendererType,
    AtomicPrescriptiveRendererType,
    LegacyDescriptiveRendererType,
    LegacyDiagnosticRendererType,
    LegacyRendererType,
    RenderedAtomicContent,
    RenderedContentBlockContainer,
    RenderedGraphContent,
    RenderedStringTemplateContent,
    renderedAtomicValueSchema,
)
from great_expectations.render.renderer.renderer import renderer
from great_expectations.render.renderer_configuration import (
    RendererConfiguration,
    RendererSchema,
    RendererTableValue,
    RendererValueType,
)
from great_expectations.render.util import (
    num_to_str,
    parse_row_condition_string_pandas_engine,
    substitute_none_for_missing,
)
from great_expectations.validator.metric_configuration import MetricConfiguration

if TYPE_CHECKING:
    from great_expectations.core import (
        ExpectationValidationResult,
    )
    from great_expectations.execution_engine import ExecutionEngine
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )
    from great_expectations.render.renderer_configuration import AddParamArgs
    from great_expectations.validator.validator import (
        ValidationDependencies,
    )

logger = logging.getLogger(__name__)
logging.captureWarnings(True)

EXPECTATION_SHORT_DESCRIPTION = (
    "Expect the Kulback-Leibler (KL) divergence (relative entropy) of the specified column "
    "with respect to the partition object to be lower than the provided threshold."
)
PARTITION_OBJECT_DESCRIPTION = "The expected partition object."

THRESHOLD_DESCRIPTION = (
    "The maximum KL divergence to for which to return success=True. If KL divergence is larger"
    "than the provided threshold, the test will return success=False."
)
INTERNAL_WEIGHT_HOLDOUT_DESCRIPTION = (
    "The amount of weight to split uniformly among zero-weighted partition bins. "
    "internal_weight_holdout provides a mechanisms to make the test less strict by "
    "assigning positive weights to values observed in the data for which "
    "the partition explicitly expected zero weight."
)
TAIL_WEIGHT_HOLDOUT_DESCRIPTION = (
    "The amount of weight to add to the tails of the histogram."
    "Tail weight holdout is split evenly between (-Infinity, min(partition_object['bins'])) "
    "and (max(partition_object['bins']), +Infinity). "
    "tail_weight_holdout provides a mechanism to make the test less strict by assigning positive "
    "weights to values observed in the data that are not present in the partition. "
    "With no tail_weight_holdout, any value observed outside "
    "the provided partition_object will cause KL divergence to rise to +Infinity. Defaults to 0."
)
BUCKETIZE_DATA_DESCRIPTION = (
    "If True, then continuous data will be bucketized before evaluation. Setting "
    "this parameter to false allows evaluation of KL divergence with a None partition object for "
    "profiling against discrete data."
)
MIN_VALUE_DESCRIPTION = "The minimum value for the column."
MAX_VALUE_DESCRIPTION = "The maximum value for the column."
SUPPORTED_DATA_SOURCES = [
    SupportedDataSources.PANDAS.value,
    SupportedDataSources.SPARK.value,
    SupportedDataSources.SQLITE.value,
    SupportedDataSources.POSTGRESQL.value,
    SupportedDataSources.MYSQL.value,
    SupportedDataSources.MSSQL.value,
    SupportedDataSources.BIGQUERY.value,
    SupportedDataSources.SNOWFLAKE.value,
    SupportedDataSources.REDSHIFT.value,
]
DATA_QUALITY_ISSUES = [DataQualityIssues.NUMERIC.value]


class ExpectColumnKLDivergenceToBeLessThan(ColumnAggregateExpectation):
    __doc__ = f"""{EXPECTATION_SHORT_DESCRIPTION}

    KL divergence compares two distributions. The higher the divergence value (relative entropy), the larger \
    the difference between the two distributions. A relative entropy of zero indicates that the data are \
    distributed identically, when binned according to the provided partition.

    In many practical contexts, choosing a value between 0.5 and 1 will provide a useful test.

    This expectation works on both categorical and continuous partitions. See notes below for details.

    ExpectColumnKlDivergenceToBeLessThan is a \
    Column Aggregate Expectation.

    Column Aggregate Expectations are one of the most common types of Expectation.
    They are evaluated for a single column, and produce an aggregate Metric, such as a mean, standard deviation, number of unique values, column type, etc.
    If that Metric meets the conditions you set, the Expectation considers that data valid.

    Args:
        column (str): \
            {COLUMN_DESCRIPTION}
        partition_object (dict or None): \
            {PARTITION_OBJECT_DESCRIPTION} See [partition_object](https://docs.greatexpectations.io/docs/reference/expectations/distributional_expectations/#partition-objects).
        threshold (float or None): \
            {THRESHOLD_DESCRIPTION}
        internal_weight_holdout (float between 0 and 1 or None): \
            {INTERNAL_WEIGHT_HOLDOUT_DESCRIPTION} With no internal_weight_holdout, any value observed in
            such a region will cause KL divergence to rise to +Infinity. Defaults to 0.
        tail_weight_holdout (float between 0 and 1 or None): \
           {TAIL_WEIGHT_HOLDOUT_DESCRIPTION}
        bucketize_data (boolean): \
            {BUCKETIZE_DATA_DESCRIPTION}
        min_value (float or None): \
            {MIN_VALUE_DESCRIPTION}
        max_value (float or None): \
            {MAX_VALUE_DESCRIPTION}

    Other Parameters:
        result_format (str or None): \
            Which output mode to use: BOOLEAN_ONLY, BASIC, COMPLETE, or SUMMARY. \
            For more detail, see [result_format](https://docs.greatexpectations.io/docs/reference/expectations/result_format).
        catch_exceptions (boolean or None): \
            If True, then catch exceptions and include them as part of the result object. \
            For more detail, see [catch_exceptions](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#catch_exceptions).
        meta (dict or None): \
            A JSON-serializable dictionary (nesting allowed) that will be included in the output without \
            modification. For more detail, see [meta](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#meta).

    Returns:
        An [ExpectationSuiteValidationResult](https://docs.greatexpectations.io/docs/terms/validation_result)

        Exact fields vary depending on the values passed to result_format, catch_exceptions, and meta.

    Notes:
        * observed_value field in the result object is customized for this expectation to be a float \
          representing the true KL divergence (relative entropy) or None if the value is calculated \
          as infinity, -infinity, or NaN
        * details.observed_partition in the result object is customized for this expectation to be a \
          dict representing the partition observed in the data
        * details.expected_partition in the result object is customized for this expectation to be a \
          dict representing the partition against which the data were compared, after applying \
          specified weight holdouts

        If the partition_object is categorical, this expectation will expect the values in column to also be \
        categorical.

        * If the column includes values that are not present in the partition, the tail_weight_holdout \
          will be equally split among those values, providing a mechanism to weaken the strictness of the \
          expectation (otherwise, relative entropy would immediately go to infinity).
        * If the partition includes values that are not present in the column, the test will simply include \
          zero weight for that value.

        If the partition_object is continuous, this expectation will discretize the values in the column according \
        to the bins specified in the partition_object, and apply the test to the resulting distribution.

        * The internal_weight_holdout and tail_weight_holdout parameters provide a mechanism to weaken the \
          expectation, since an expected weight of zero would drive relative entropy to be infinite if any data \
          are observed in that interval.
        * If internal_weight_holdout is specified, that value will be distributed equally among any intervals \
          with weight zero in the partition_object.
        * If tail_weight_holdout is specified, that value will be appended to the tails of the bins \
          ((-Infinity, min(bins)) and (max(bins), Infinity).

        If relative entropy/kl divergence goes to infinity for any of the reasons mentioned above, the observed \
        value will be set to None. This is because inf, -inf, Nan, are not json serializable and cause some json \
        parsers to crash when encountered. The python None token will be serialized to null in json.

    Supported Data Sources:
        [{SUPPORTED_DATA_SOURCES[0]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[1]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[2]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[3]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[4]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[5]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[6]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[7]}](https://docs.greatexpectations.io/docs/application_integration_support/)

    Data Quality Issues:
        {DATA_QUALITY_ISSUES[0]}

    Example Data:
                test
            0 	"A"
            1 	"A"
            2 	"A"
            3   "A"
            4   "A"
            5   "B"
            6   "B"
            7   "B"
            8   "C"
            9   "C"

    Code Examples:
        Passing Case:
            Input:
                ExpectColumnKLDivergenceToBeLessThan(
                    column="test",
                    partition_object={{"weights": [0.5, 0.3, 0.2], "values": ["A", "B", "C"]}},
                    threshold=0.1
            )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "observed_value": 0.0,
                    "details": {{
                      "observed_partition": {{
                        "values": [
                          "A",
                          "B",
                          "C"
                        ],
                        "weights": [
                          0.5,
                          0.3,
                          0.2
                        ]
                      }},
                      "expected_partition": {{
                        "values": [
                          "A",
                          "B",
                          "C"
                        ],
                        "weights": [
                          0.5,
                          0.3,
                          0.2
                        ]
                      }}
                    }}
                  }},
                  "meta": {{}},
                  "success": true
                }}

        Failing Case:
            Input:
                ExpectColumnKLDivergenceToBeLessThan(
                    column="test",
                    partition_object={{"weights": [0.3333333333333333, 0.3333333333333333, 0.3333333333333333], "values": ["A", "B", "C"]}},
                    threshold=0.01
                )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "observed_value": 0.06895927460353621,
                    "details": {{
                      "observed_partition": {{
                        "values": [
                          "A",
                          "B",
                          "C"
                        ],
                        "weights": [
                          0.5,
                          0.3,
                          0.2
                        ]
                      }},
                      "expected_partition": {{
                        "values": [
                          "A",
                          "B",
                          "C"
                        ],
                        "weights": [
                          0.3333333333333333,
                          0.3333333333333333,
                          0.3333333333333333
                        ]
                      }}
                    }}
                  }},
                  "meta": {{}},
                  "success": false
                }}
    """  # noqa: E501 # FIXME CoP

    partition_object: Union[dict, SuiteParameterDict, None] = pydantic.Field(
        description=PARTITION_OBJECT_DESCRIPTION
    )
    threshold: Union[float, SuiteParameterDict, None] = pydantic.Field(
        description=THRESHOLD_DESCRIPTION
    )
    internal_weight_holdout: Union[float, SuiteParameterDict, None] = pydantic.Field(
        default=0, ge=0, le=1, description=INTERNAL_WEIGHT_HOLDOUT_DESCRIPTION
    )
    tail_weight_holdout: Union[float, SuiteParameterDict, None] = pydantic.Field(
        default=0, ge=0, le=1, description=TAIL_WEIGHT_HOLDOUT_DESCRIPTION
    )
    bucketize_data: Union[bool, SuiteParameterDict] = pydantic.Field(
        default=True, description=BUCKETIZE_DATA_DESCRIPTION
    )
    min_value: Optional[Comparable] = pydantic.Field(
        default=None, description=MIN_VALUE_DESCRIPTION
    )
    max_value: Optional[Comparable] = pydantic.Field(
        default=None, description=MAX_VALUE_DESCRIPTION
    )

    # This dictionary contains metadata for display in the public gallery
    library_metadata: ClassVar[Dict[str, Union[str, list, bool]]] = {
        "maturity": "production",
        "tags": [
            "core expectation",
            "column aggregate expectation",
            "distributional expectation",
        ],
        "contributors": ["@great_expectations"],
        "requirements": [],
        "has_full_test_suite": True,
        "manually_reviewed_code": True,
    }

    _library_metadata = library_metadata

    success_keys = (
        "partition_object",
        "threshold",
        "tail_weight_holdout",
        "internal_weight_holdout",
        "bucketize_data",
    )
    args_keys = (
        "column",
        "partition_object",
        "threshold",
    )

    class Config:
        title = "Expect column KL divergence to be less than"

        @staticmethod
        def schema_extra(
            schema: Dict[str, Any], model: Type[ExpectColumnKLDivergenceToBeLessThan]
        ) -> None:
            ColumnAggregateExpectation.Config.schema_extra(schema, model)
            schema["properties"]["metadata"]["properties"].update(
                {
                    "data_quality_issues": {
                        "title": "Data Quality Issues",
                        "type": "array",
                        "const": DATA_QUALITY_ISSUES,
                    },
                    "library_metadata": {
                        "title": "Library Metadata",
                        "type": "object",
                        "const": model._library_metadata,
                    },
                    "short_description": {
                        "title": "Short Description",
                        "type": "string",
                        "const": EXPECTATION_SHORT_DESCRIPTION,
                    },
                    "supported_data_sources": {
                        "title": "Supported Data Sources",
                        "type": "array",
                        "const": SUPPORTED_DATA_SOURCES,
                    },
                }
            )

    def get_validation_dependencies(
        self,
        execution_engine: Optional[ExecutionEngine] = None,
        runtime_configuration: Optional[dict] = None,
    ) -> ValidationDependencies:
        validation_dependencies: ValidationDependencies = super().get_validation_dependencies(
            execution_engine, runtime_configuration
        )
        configuration = self.configuration
        partition_object = configuration.kwargs.get("partition_object")
        domain_kwargs = configuration.get_domain_kwargs()
        is_categorical = None
        bins = None
        if partition_object is None:
            if configuration.kwargs.get(
                "bucketize_data", self._get_default_value("bucketize_data")
            ):
                is_categorical = False
                partition_metric_configuration = MetricConfiguration(
                    metric_name="column.partition",
                    metric_domain_kwargs=domain_kwargs,
                    metric_value_kwargs={
                        "bins": "auto",
                        "allow_relative_error": False,
                    },
                )
                #
                # Note: 20201116 - JPC - the execution engine doesn't provide capability to evaluate
                # validation_dependencies, so we use a validator
                #
                from great_expectations.validator.metrics_calculator import (
                    MetricsCalculator,
                )

                metrics_calculator = MetricsCalculator(
                    execution_engine=execution_engine,
                    show_progress_bars=True,
                )
                resolved_metrics, _ = metrics_calculator.compute_metrics(
                    metric_configurations=[partition_metric_configuration],
                    runtime_configuration=None,
                    min_graph_edges_pbar_enable=0,
                )

                bins = resolved_metrics[partition_metric_configuration.id]
                hist_metric_configuration = MetricConfiguration(
                    metric_name="column.histogram",
                    metric_domain_kwargs=domain_kwargs,
                    metric_value_kwargs={
                        "bins": tuple(bins),
                    },
                )
                nonnull_configuration = MetricConfiguration(
                    metric_name="column_values.nonnull.count",
                    metric_domain_kwargs=domain_kwargs,
                    metric_value_kwargs=None,
                )
                #
                # NOTE 20201117 - JPC - Would prefer not to include partition_metric_configuration here,  # noqa: E501 # FIXME CoP
                # since we have already evaluated it, and its result is in the kwargs for the histogram.  # noqa: E501 # FIXME CoP
                # However, currently the validation_dependencies' configurations are not passed to the _validate method  # noqa: E501 # FIXME CoP
                #
                validation_dependencies.set_metric_configuration(
                    metric_name="column.partition",
                    metric_configuration=partition_metric_configuration,
                )
                validation_dependencies.set_metric_configuration(
                    metric_name="column.histogram",
                    metric_configuration=hist_metric_configuration,
                )
                validation_dependencies.set_metric_configuration(
                    metric_name="column_values.nonnull.count",
                    metric_configuration=nonnull_configuration,
                )
            else:
                is_categorical = True
                counts_configuration = MetricConfiguration(
                    metric_name="column.value_counts",
                    metric_domain_kwargs=domain_kwargs,
                    metric_value_kwargs={
                        "sort": "value",
                    },
                )
                nonnull_configuration = MetricConfiguration(
                    metric_name="column_values.nonnull.count",
                    metric_domain_kwargs=domain_kwargs,
                )
                validation_dependencies.set_metric_configuration(
                    metric_name="column.value_counts",
                    metric_configuration=counts_configuration,
                )
                validation_dependencies.set_metric_configuration(
                    metric_name="column_values.nonnull.count",
                    metric_configuration=nonnull_configuration,
                )
        if is_categorical is True or is_valid_categorical_partition_object(partition_object):
            validation_dependencies.set_metric_configuration(
                metric_name="column.value_counts",
                metric_configuration=MetricConfiguration(
                    metric_name="column.value_counts",
                    metric_domain_kwargs=domain_kwargs,
                    metric_value_kwargs={
                        "sort": "value",
                    },
                ),
            )
            validation_dependencies.set_metric_configuration(
                metric_name="column_values.nonnull.count",
                metric_configuration=MetricConfiguration(
                    metric_name="column_values.nonnull.count",
                    metric_domain_kwargs=domain_kwargs,
                    metric_value_kwargs=None,
                ),
            )
        else:
            if (
                bins is None
            ):  # if the user did not supply a partition_object, so we just computed it
                if not is_valid_partition_object(partition_object):
                    raise ValueError("Invalid partition_object provided")  # noqa: TRY003 # FIXME CoP
                bins = partition_object["bins"]

            hist_metric_configuration = MetricConfiguration(
                metric_name="column.histogram",
                metric_domain_kwargs=domain_kwargs,
                metric_value_kwargs={
                    "bins": bins,
                },
            )
            validation_dependencies.set_metric_configuration(
                metric_name="column.histogram",
                metric_configuration=hist_metric_configuration,
            )
            nonnull_configuration = MetricConfiguration(
                metric_name="column_values.nonnull.count",
                metric_domain_kwargs=domain_kwargs,
                metric_value_kwargs=None,
            )
            validation_dependencies.set_metric_configuration(
                metric_name="column_values.nonnull.count",
                metric_configuration=nonnull_configuration,
            )
            below_partition = MetricConfiguration(
                metric_name="column_values.between.count",
                metric_domain_kwargs=domain_kwargs,
                metric_value_kwargs={"max_value": bins[0], "strict_max": True},
            )
            validation_dependencies.set_metric_configuration(
                metric_name="below_partition", metric_configuration=below_partition
            )
            above_partition = MetricConfiguration(
                metric_name="column_values.between.count",
                metric_domain_kwargs=domain_kwargs,
                metric_value_kwargs={"min_value": bins[-1], "strict_min": True},
            )
            validation_dependencies.set_metric_configuration(
                metric_name="above_partition", metric_configuration=above_partition
            )

        return validation_dependencies

    def _validate(  # noqa: C901, PLR0912, PLR0915 # FIXME CoP
        self,
        metrics: Dict,
        runtime_configuration: Optional[dict] = None,
        execution_engine: Optional[ExecutionEngine] = None,
    ):
        configuration = self.configuration
        bucketize_data = configuration.kwargs.get(
            "bucketize_data", self._get_default_value("bucketize_data")
        )
        partition_object = configuration.kwargs.get(
            "partition_object", self._get_default_value("partition_object")
        )
        threshold = configuration.kwargs.get("threshold", self._get_default_value("threshold"))
        tail_weight_holdout = configuration.kwargs.get(
            "tail_weight_holdout", self._get_default_value("tail_weight_holdout")
        )
        internal_weight_holdout = configuration.kwargs.get(
            "internal_weight_holdout",
            self._get_default_value("internal_weight_holdout"),
        )
        if partition_object is None:
            if bucketize_data:
                # in this case, we have requested a partition, histogram using said partition, and nonnull count  # noqa: E501 # FIXME CoP
                bins = list(metrics["column.partition"])
                weights = list(
                    np.array(metrics["column.histogram"]) / metrics["column_values.nonnull.count"]
                )
                tail_weights = (1 - sum(weights)) / 2
                partition_object = {
                    "bins": bins,
                    "weights": weights,
                    "tail_weights": [tail_weights, tail_weights],
                }
            else:
                partition_object = {
                    "values": list(metrics["column.value_counts"].index),
                    "weights": list(
                        np.array(metrics["column.value_counts"])
                        / metrics["column_values.nonnull.count"]
                    ),
                }

        if not is_valid_partition_object(partition_object):
            raise ValueError("Invalid partition object.")  # noqa: TRY003 # FIXME CoP

        if threshold is not None and ((not isinstance(threshold, (int, float))) or (threshold < 0)):
            raise ValueError("Threshold must be specified, greater than or equal to zero.")  # noqa: TRY003 # FIXME CoP

        if (
            (not isinstance(tail_weight_holdout, (int, float)))
            or (tail_weight_holdout < 0)
            or (tail_weight_holdout > 1)
        ):
            raise ValueError("tail_weight_holdout must be between zero and one.")  # noqa: TRY003 # FIXME CoP

        if (
            (not isinstance(internal_weight_holdout, (int, float)))
            or (internal_weight_holdout < 0)
            or (internal_weight_holdout > 1)
        ):
            raise ValueError("internal_weight_holdout must be between zero and one.")  # noqa: TRY003 # FIXME CoP

        if tail_weight_holdout != 0 and "tail_weights" in partition_object:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                "tail_weight_holdout must be 0 when using tail_weights in partition object"
            )

        # TODO: add checks for duplicate values in is_valid_categorical_partition_object
        if is_valid_categorical_partition_object(partition_object):
            if internal_weight_holdout > 0:
                raise ValueError("Internal weight holdout cannot be used for discrete data.")  # noqa: TRY003 # FIXME CoP

            # Data are expected to be discrete, use value_counts
            observed_weights = (
                metrics["column.value_counts"] / metrics["column_values.nonnull.count"]
            )
            expected_weights = pd.Series(
                partition_object["weights"],
                index=partition_object["values"],
                name="expected",
            )
            # Sort not available before pandas 0.23.0
            # test_df = pd.concat([expected_weights, observed_weights], axis=1, sort=True)
            test_df = pd.concat([expected_weights, observed_weights], axis=1)

            na_counts = test_df.isnull().sum()

            # Handle NaN: if we expected something that's not there, it's just not there.
            pk = test_df["count"].fillna(0)
            # Handle NaN: if something's there that was not expected,
            # substitute the relevant value for tail_weight_holdout
            if na_counts["expected"] > 0:
                # Scale existing expected values
                test_df["expected"] *= 1 - tail_weight_holdout
                # Fill NAs with holdout.
                qk = test_df["expected"].fillna(tail_weight_holdout / na_counts["expected"])
            else:
                qk = test_df["expected"]

            kl_divergence = stats.entropy(pk, qk)

            if np.isinf(kl_divergence) or np.isnan(kl_divergence):
                observed_value = None
            else:
                observed_value = kl_divergence

            if threshold is None:
                success = True
            else:
                success = kl_divergence <= threshold

            return_obj = {
                "success": success,
                "result": {
                    "observed_value": observed_value,
                    "details": {
                        "observed_partition": {
                            "values": test_df.index.tolist(),
                            "weights": pk.tolist(),
                        },
                        "expected_partition": {
                            "values": test_df.index.tolist(),
                            "weights": qk.tolist(),
                        },
                    },
                },
            }

        else:
            # Data are expected to be continuous; discretize first
            if bucketize_data is False:
                raise ValueError(  # noqa: TRY003 # FIXME CoP
                    "KL Divergence cannot be computed with a continuous partition object and the bucketize_data "  # noqa: E501 # FIXME CoP
                    "parameter set to false."
                )
            # Build the histogram first using expected bins so that the largest bin is >=
            nonnull_count = metrics["column_values.nonnull.count"]
            hist = np.array(metrics["column.histogram"])

            # Add in the frequencies observed above or below the provided partition
            below_partition = metrics["below_partition"]
            above_partition = metrics["above_partition"]

            # Observed Weights is just the histogram values divided by the total number of observations  # noqa: E501 # FIXME CoP
            observed_weights = hist / nonnull_count

            # Adjust expected_weights to account for tail_weight and internal_weight
            if "tail_weights" in partition_object:
                partition_tail_weight_holdout = np.sum(partition_object["tail_weights"])
            else:
                partition_tail_weight_holdout = 0  # noqa: F841 # unused

            expected_weights = np.array(partition_object["weights"]) * (
                1 - tail_weight_holdout - internal_weight_holdout
            )

            # Assign internal weight holdout values if applicable
            if internal_weight_holdout > 0:
                zero_count = len(expected_weights) - np.count_nonzero(expected_weights)
                if zero_count > 0:
                    for index, value in enumerate(expected_weights):
                        if value == 0:
                            expected_weights[index] = internal_weight_holdout / zero_count

            # Assign tail weight holdout if applicable
            # We need to check cases to only add tail weight holdout if it makes sense based on the provided partition.  # noqa: E501 # FIXME CoP
            if (partition_object["bins"][0] == -np.inf) and (
                partition_object["bins"][-1]
            ) == np.inf:
                if tail_weight_holdout > 0:
                    raise ValueError(  # noqa: TRY003 # FIXME CoP
                        "tail_weight_holdout cannot be used for partitions with infinite endpoints."
                    )
                if "tail_weights" in partition_object:
                    raise ValueError(  # noqa: TRY003 # FIXME CoP
                        "There can be no tail weights for partitions with one or both endpoints at infinity"  # noqa: E501 # FIXME CoP
                    )

                # Remove -inf and inf
                expected_bins = partition_object["bins"][1:-1]

                comb_expected_weights = expected_weights
                # Set aside tail weights
                expected_tail_weights = np.concatenate(
                    ([expected_weights[0]], [expected_weights[-1]])
                )
                # Remove tail weights
                expected_weights = expected_weights[1:-1]

                comb_observed_weights = observed_weights
                # Set aside tail weights
                observed_tail_weights = np.concatenate(
                    ([observed_weights[0]], [observed_weights[-1]])
                )
                # Remove tail weights
                observed_weights = observed_weights[1:-1]

            elif partition_object["bins"][0] == -np.inf:
                if "tail_weights" in partition_object:
                    raise ValueError(  # noqa: TRY003 # FIXME CoP
                        "There can be no tail weights for partitions with one or both endpoints at infinity"  # noqa: E501 # FIXME CoP
                    )

                # Remove -inf
                expected_bins = partition_object["bins"][1:]

                comb_expected_weights = np.concatenate((expected_weights, [tail_weight_holdout]))
                # Set aside left tail weight and holdout
                expected_tail_weights = np.concatenate(
                    ([expected_weights[0]], [tail_weight_holdout])
                )
                # Remove left tail weight from main expected_weights
                expected_weights = expected_weights[1:]

                comb_observed_weights = np.concatenate(
                    (
                        observed_weights,
                        [above_partition / nonnull_count],
                    )
                )
                # Set aside left tail weight and above partition weight
                observed_tail_weights = np.concatenate(
                    (
                        [observed_weights[0]],
                        [above_partition / nonnull_count],
                    )
                )
                # Remove left tail weight from main observed_weights
                observed_weights = observed_weights[1:]

            elif partition_object["bins"][-1] == np.inf:
                if "tail_weights" in partition_object:
                    raise ValueError(  # noqa: TRY003 # FIXME CoP
                        "There can be no tail weights for partitions with one or both endpoints at infinity"  # noqa: E501 # FIXME CoP
                    )

                # Remove inf
                expected_bins = partition_object["bins"][:-1]

                comb_expected_weights = np.concatenate(([tail_weight_holdout], expected_weights))
                # Set aside right tail weight and holdout
                expected_tail_weights = np.concatenate(
                    ([tail_weight_holdout], [expected_weights[-1]])
                )
                # Remove right tail weight from main expected_weights
                expected_weights = expected_weights[:-1]

                comb_observed_weights = np.concatenate(
                    (
                        [below_partition / nonnull_count],
                        observed_weights,
                    )
                )
                # Set aside right tail weight and below partition weight
                observed_tail_weights = np.concatenate(
                    (
                        [below_partition / nonnull_count],
                        [observed_weights[-1]],
                    )
                )
                # Remove right tail weight from main observed_weights
                observed_weights = observed_weights[:-1]
            else:
                # No need to remove -inf or inf
                expected_bins = partition_object["bins"]

                if "tail_weights" in partition_object:
                    tail_weights = partition_object["tail_weights"]
                    # Tack on tail weights
                    comb_expected_weights = np.concatenate(
                        ([tail_weights[0]], expected_weights, [tail_weights[1]])
                    )
                    # Tail weights are just tail_weights
                    expected_tail_weights = np.array(tail_weights)
                else:
                    comb_expected_weights = np.concatenate(
                        (
                            [tail_weight_holdout / 2],
                            expected_weights,
                            [tail_weight_holdout / 2],
                        )
                    )
                    # Tail weights are just tail_weight holdout divided equally to both tails
                    expected_tail_weights = np.concatenate(
                        ([tail_weight_holdout / 2], [tail_weight_holdout / 2])
                    )

                comb_observed_weights = np.concatenate(
                    (
                        [below_partition / nonnull_count],
                        observed_weights,
                        [above_partition / nonnull_count],
                    )
                )
                # Tail weights are just the counts on either side of the partition
                observed_tail_weights = (
                    np.concatenate(([below_partition], [above_partition])) / nonnull_count
                )

                # Main expected_weights and main observed weights had no tail_weights, so nothing needs to be removed.  # noqa: E501 # FIXME CoP

            # TODO: VERIFY THAT THIS STILL WORKS BASED ON CHANGE TO HIST
            # comb_expected_weights = np.array(comb_expected_weights).astype(float)
            # comb_observed_weights = np.array(comb_observed_weights).astype(float)

            kl_divergence = stats.entropy(comb_observed_weights, comb_expected_weights)

            if np.isinf(kl_divergence) or np.isnan(kl_divergence):
                observed_value = None
            else:
                observed_value = kl_divergence

            if threshold is None:
                success = True
            else:
                success = kl_divergence <= threshold

            return_obj = {
                "success": success,
                "result": {
                    "observed_value": observed_value,
                    "details": {
                        "observed_partition": {
                            # return expected_bins, since we used those bins to compute the observed_weights  # noqa: E501 # FIXME CoP
                            "bins": expected_bins,
                            "weights": observed_weights.tolist(),
                            "tail_weights": observed_tail_weights.tolist(),
                        },
                        "expected_partition": {
                            "bins": expected_bins,
                            "weights": expected_weights.tolist(),
                            "tail_weights": expected_tail_weights.tolist(),
                        },
                    },
                },
            }

        return return_obj

    @classmethod
    def _get_kl_divergence_chart(  # noqa: C901 #  13
        cls, partition_object, header=None
    ):
        weights = partition_object["weights"]

        if len(weights) > 60:  # noqa: PLR2004 # FIXME CoP
            expected_distribution = cls._get_kl_divergence_partition_object_table(
                partition_object, header=header
            )
        else:
            chart_pixel_width = (len(weights) / 60.0) * 500
            chart_pixel_width = max(chart_pixel_width, 250)
            chart_container_col_width = round((len(weights) / 60.0) * 6)
            if chart_container_col_width < 4:  # noqa: PLR2004 # FIXME CoP
                chart_container_col_width = 4
            elif chart_container_col_width >= 5:  # noqa: PLR2004 # FIXME CoP
                chart_container_col_width = 6
            elif chart_container_col_width >= 4:  # noqa: PLR2004 # FIXME CoP
                chart_container_col_width = 5

            mark_bar_args = {}
            if len(weights) == 1:
                mark_bar_args["size"] = 20

            if partition_object.get("bins"):
                bins = partition_object["bins"]
                bins_x1 = [round(value, 1) for value in bins[:-1]]
                bins_x2 = [round(value, 1) for value in bins[1:]]

                df = pd.DataFrame(
                    {
                        "bin_min": bins_x1,
                        "bin_max": bins_x2,
                        "fraction": weights,
                    }
                )

                bars = (
                    alt.Chart(df)
                    .mark_bar()
                    .encode(
                        x="bin_min:O",
                        x2="bin_max:O",
                        y="fraction:Q",
                        tooltip=["bin_min", "bin_max", "fraction"],
                    )
                    .properties(width=chart_pixel_width, height=400, autosize="fit")
                )

                chart = bars.to_dict()
            elif partition_object.get("values"):
                is_boolean_list = all(
                    isinstance(value, bool) for value in partition_object["values"]
                )
                if is_boolean_list:
                    values = [str(value) for value in partition_object["values"]]
                else:
                    values = partition_object["values"]

                df = pd.DataFrame({"values": values, "fraction": weights})

                bars = (
                    alt.Chart(df)
                    .mark_bar()
                    .encode(x="values:N", y="fraction:Q", tooltip=["values", "fraction"])
                    .properties(width=chart_pixel_width, height=400, autosize="fit")
                )
                chart = bars.to_dict()

            if header:
                expected_distribution = RenderedGraphContent(
                    **{
                        "content_block_type": "graph",
                        "graph": chart,
                        "header": header,
                        "styling": {
                            "classes": [
                                f"col-{chart_container_col_width!s}",
                                "mt-2",
                                "pl-1",
                                "pr-1",
                            ],
                            "parent": {"styles": {"list-style-type": "none"}},
                        },
                    }
                )
            else:
                expected_distribution = RenderedGraphContent(
                    **{
                        "content_block_type": "graph",
                        "graph": chart,
                        "styling": {
                            "classes": [
                                f"col-{chart_container_col_width!s}",
                                "mt-2",
                                "pl-1",
                                "pr-1",
                            ],
                            "parent": {"styles": {"list-style-type": "none"}},
                        },
                    }
                )
        return expected_distribution

    @classmethod
    def _atomic_kl_divergence_chart_template(cls, partition_object: dict) -> tuple:
        weights = partition_object.get("weights", [])

        chart_pixel_width = (len(weights) / 60.0) * 500
        chart_pixel_width = max(chart_pixel_width, 250)
        chart_container_col_width = round((len(weights) / 60.0) * 6)
        if chart_container_col_width < 4:  # noqa: PLR2004 # FIXME CoP
            chart_container_col_width = 4
        elif chart_container_col_width >= 5:  # noqa: PLR2004 # FIXME CoP
            chart_container_col_width = 6
        elif chart_container_col_width >= 4:  # noqa: PLR2004 # FIXME CoP
            chart_container_col_width = 5

        mark_bar_args = {}
        if len(weights) == 1:
            mark_bar_args["size"] = 20

        chart = {}
        if partition_object.get("bins"):
            bins = partition_object["bins"]
            bins_x1 = [round(value, 1) for value in bins[:-1]]
            bins_x2 = [round(value, 1) for value in bins[1:]]

            df = pd.DataFrame(
                {
                    "bin_min": bins_x1,
                    "bin_max": bins_x2,
                    "fraction": weights,
                }
            )

            bars = (
                alt.Chart(df)
                .mark_bar()
                .encode(
                    x="bin_min:O",
                    x2="bin_max:O",
                    y="fraction:Q",
                    tooltip=["bin_min", "bin_max", "fraction"],
                )
                .properties(width=chart_pixel_width, height=400, autosize="fit")
            )

            chart = bars.to_dict()
        elif partition_object.get("values"):
            is_boolean_list = all(isinstance(value, bool) for value in partition_object["values"])
            if is_boolean_list:
                values = [str(value) for value in partition_object["values"]]
            else:
                values = partition_object["values"]

            df = pd.DataFrame({"values": values, "fraction": weights})

            bars = (
                alt.Chart(df)
                .mark_bar()
                .encode(x="values:N", y="fraction:Q", tooltip=["values", "fraction"])
                .properties(width=chart_pixel_width, height=400, autosize="fit")
            )
            chart = bars.to_dict()

        return chart, chart_container_col_width

    @classmethod
    def _get_kl_divergence_partition_object_table(cls, partition_object, header=None):
        table_rows = []
        fractions = partition_object["weights"]

        if partition_object.get("bins"):
            bins = partition_object["bins"]

            for idx, fraction in enumerate(fractions):
                if idx == len(fractions) - 1:
                    table_rows.append(
                        [
                            f"[{num_to_str(bins[idx])} - {num_to_str(bins[idx + 1])}]",
                            num_to_str(fraction),
                        ]
                    )
                else:
                    table_rows.append(
                        [
                            f"[{num_to_str(bins[idx])} - {num_to_str(bins[idx + 1])})",
                            num_to_str(fraction),
                        ]
                    )
        else:
            values = partition_object["values"]
            table_rows = [[value, num_to_str(fractions[idx])] for idx, value in enumerate(values)]

        if header:
            return {
                "content_block_type": "table",
                "header": header,
                "header_row": (
                    ["Interval", "Fraction"]
                    if partition_object.get("bins")
                    else ["Value", "Fraction"]
                ),
                "table": table_rows,
                "styling": {
                    "classes": ["table-responsive"],
                    "body": {
                        "classes": [
                            "table",
                            "table-sm",
                            "table-bordered",
                            "mt-2",
                            "mb-2",
                        ],
                    },
                    "parent": {
                        "classes": ["show-scrollbars", "p-2"],
                        "styles": {
                            "list-style-type": "none",
                            "overflow": "auto",
                            "max-height": "80vh",
                        },
                    },
                },
            }
        else:
            return {
                "content_block_type": "table",
                "header_row": (
                    ["Interval", "Fraction"]
                    if partition_object.get("bins")
                    else ["Value", "Fraction"]
                ),
                "table": table_rows,
                "styling": {
                    "classes": ["table-responsive"],
                    "body": {
                        "classes": [
                            "table",
                            "table-sm",
                            "table-bordered",
                            "mt-2",
                            "mb-2",
                        ],
                    },
                    "parent": {
                        "classes": ["show-scrollbars", "p-2"],
                        "styles": {
                            "list-style-type": "none",
                            "overflow": "auto",
                            "max-height": "80vh",
                        },
                    },
                },
            }

    @classmethod
    def _atomic_partition_object_table_template(cls, partition_object: dict):
        table_rows = []
        fractions = partition_object["weights"]

        if partition_object.get("bins"):
            bins = partition_object["bins"]

            for idx, fraction in enumerate(fractions):
                interval_start = num_to_str(bins[idx])
                interval_end = num_to_str(bins[idx + 1])
                interval_closing_symbol = "]" if idx == (len(fractions) - 1) else ")"
                table_rows.append(
                    [
                        RendererTableValue(
                            schema=RendererSchema(type=RendererValueType.STRING),
                            value=f"[{interval_start} - {interval_end}{interval_closing_symbol}",
                        ),
                        RendererTableValue(
                            schema=RendererSchema(type=RendererValueType.STRING),
                            value=num_to_str(fraction),
                        ),
                    ]
                )
        else:
            values = partition_object["values"]
            table_rows = [
                [
                    RendererTableValue(
                        schema=RendererSchema(type=RendererValueType.STRING),
                        value=str(value),
                    ),
                    RendererTableValue(
                        schema=RendererSchema(type=RendererValueType.STRING),
                        value=num_to_str(fractions[idx]),
                    ),
                ]
                for idx, value in enumerate(values)
            ]

        interval_or_value = "Interval" if partition_object.get("bins") else "Value"

        header_row = [
            RendererTableValue(
                schema=RendererSchema(type=RendererValueType.STRING),
                value=interval_or_value,
            ),
            RendererTableValue(
                schema=RendererSchema(type=RendererValueType.STRING), value="Fraction"
            ),
        ]

        return header_row, table_rows

    @classmethod
    def _prescriptive_template(
        cls,
        renderer_configuration: RendererConfiguration,
    ) -> RendererConfiguration:
        add_param_args: AddParamArgs = (
            ("column", RendererValueType.STRING),
            ("mostly", RendererValueType.NUMBER),
            ("threshold", RendererValueType.NUMBER),
        )
        for name, param_type in add_param_args:
            renderer_configuration.add_param(name=name, param_type=param_type)

        expected_partition_object = renderer_configuration.kwargs.get("partition_object", {})
        weights = expected_partition_object.get("weights", [])

        # generate template string for header
        if not expected_partition_object:
            template_str = "can match any distribution."
        else:
            template_str = (
                "Kullback-Leibler (KL) divergence with respect to the following distribution must be "  # noqa: E501 # FIXME CoP
                "lower than $threshold."
            )

        # optionally, add column name
        if renderer_configuration.include_column_name:
            template_str = f"$column {template_str}"

        # generate table or chart depending on number of weights
        if len(weights) > 60:  # noqa: PLR2004 # FIXME CoP
            (
                renderer_configuration.header_row,
                renderer_configuration.table,
            ) = cls._atomic_partition_object_table_template(
                partition_object=expected_partition_object
            )
        else:
            renderer_configuration.graph, _ = cls._atomic_kl_divergence_chart_template(
                partition_object=expected_partition_object
            )

        renderer_configuration.template_str = template_str

        return renderer_configuration

    @classmethod
    @renderer(renderer_type=AtomicPrescriptiveRendererType.SUMMARY)
    @render_suite_parameter_string
    def _prescriptive_summary(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
    ) -> RenderedAtomicContent:
        renderer_configuration = RendererConfiguration(
            configuration=configuration,
            result=result,
            runtime_configuration=runtime_configuration,
        )
        renderer_configuration = cls._prescriptive_template(
            renderer_configuration=renderer_configuration
        )

        if renderer_configuration.graph:
            value_obj = renderedAtomicValueSchema.load(
                {
                    "header": {
                        "schema": {"type": "StringValueType"},
                        "value": {
                            "template": renderer_configuration.template_str,
                            "params": renderer_configuration.params.dict(),
                        },
                    },
                    "graph": renderer_configuration.graph,
                    "meta_notes": renderer_configuration.meta_notes,
                    "schema": {"type": "GraphType"},
                }
            )
            value_type = "GraphType"
        else:
            header_row = [value.dict() for value in renderer_configuration.header_row]
            table = []
            for row in renderer_configuration.table:
                table.append([value.dict() for value in row])
            value_obj = renderedAtomicValueSchema.load(
                {
                    "header": {
                        "schema": {"type": "StringValueType"},
                        "value": {
                            "template": renderer_configuration.template_str,
                            "params": renderer_configuration.params.dict(),
                        },
                    },
                    "header_row": header_row,
                    "table": table,
                    "meta_notes": renderer_configuration.meta_notes,
                    "schema": {"type": "TableType"},
                }
            )
            value_type = "TableType"

        return RenderedAtomicContent(
            name="atomic.prescriptive.summary",
            value=value_obj,
            value_type=value_type,
        )

    @classmethod
    @renderer(renderer_type=LegacyRendererType.PRESCRIPTIVE)
    @render_suite_parameter_string
    def _prescriptive_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        runtime_configuration = runtime_configuration or {}
        include_column_name = runtime_configuration.get("include_column_name") is not False
        _ = runtime_configuration.get("styling")
        params = substitute_none_for_missing(
            configuration.kwargs,
            [
                "column",
                "partition_object",
                "threshold",
                "row_condition",
                "condition_parser",
            ],
        )

        expected_distribution = None
        if not params.get("partition_object"):
            template_str = "can match any distribution."
        else:
            template_str = (
                "Kullback-Leibler (KL) divergence with respect to the following distribution must be "  # noqa: E501 # FIXME CoP
                "lower than $threshold."
            )
            expected_distribution = cls._get_kl_divergence_chart(params.get("partition_object"))

        if include_column_name:
            template_str = f"$column {template_str}"

        if params["row_condition"] is not None:
            (
                conditional_template_str,
                conditional_params,
            ) = parse_row_condition_string_pandas_engine(params["row_condition"])
            template_str = f"{conditional_template_str}, then {template_str}"
            params.update(conditional_params)

        expectation_string_obj = {
            "content_block_type": "string_template",
            "string_template": {"template": template_str, "params": params},
        }

        if expected_distribution:
            return [expectation_string_obj, expected_distribution]
        else:
            return [expectation_string_obj]

    @classmethod
    def _atomic_diagnostic_observed_value_template(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
    ):
        observed_partition_object = result.result.get("details", {}).get("observed_partition", {})
        weights = observed_partition_object.get("weights", [])

        observed_value = (
            num_to_str(result.result.get("observed_value"))
            if result.result.get("observed_value")
            else result.result.get("observed_value")
        )
        header_template_str = "KL Divergence: $observed_value"
        header_params_with_json_schema = {
            "observed_value": {
                "schema": {"type": "string"},
                "value": (
                    str(observed_value) if observed_value else "None (-infinity, infinity, or NaN)"
                ),
            }
        }

        chart = None
        chart_container_col_width = None
        distribution_table_header_row = None
        distribution_table_rows = None

        if len(weights) > 60:  # noqa: PLR2004 # FIXME CoP
            (
                distribution_table_header_row,
                distribution_table_rows,
            ) = cls._atomic_partition_object_table_template(
                partition_object=observed_partition_object
            )
        else:
            chart, chart_container_col_width = cls._atomic_kl_divergence_chart_template(
                partition_object=observed_partition_object
            )

        return (
            header_template_str,
            header_params_with_json_schema,
            chart,
            chart_container_col_width,
            distribution_table_header_row,
            distribution_table_rows,
        )

    @classmethod
    @renderer(renderer_type=AtomicDiagnosticRendererType.OBSERVED_VALUE)
    def _atomic_diagnostic_observed_value(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
    ):
        if not result.result.get("details"):
            value_obj = renderedAtomicValueSchema.load(
                {
                    "template": "--",
                    "params": {},
                    "schema": {"type": "StringValueType"},
                }
            )
            return RenderedAtomicContent(
                name="atomic.diagnostic.observed_value",
                value=value_obj,
                value_type="StringValueType",
            )

        (
            header_template_str,
            header_params_with_json_schema,
            chart,
            _chart_container_col_width,
            distribution_table_header_row,
            distribution_table_rows,
        ) = cls._atomic_diagnostic_observed_value_template(
            configuration,
            result,
            runtime_configuration,
        )

        if chart is not None:
            value_obj = renderedAtomicValueSchema.load(
                {
                    "header": {
                        "schema": {"type": "StringValueType"},
                        "value": {
                            "template": header_template_str,
                            "params": header_params_with_json_schema,
                        },
                    },
                    "graph": chart,
                    "schema": {"type": "GraphType"},
                }
            )
            value_type = "GraphType"
        else:
            value_obj = renderedAtomicValueSchema.load(
                {
                    "header": {
                        "schema": {"type": "StringValueType"},
                        "value": {
                            "template": header_template_str,
                            "params": header_params_with_json_schema,
                        },
                    },
                    "header_row": distribution_table_header_row,
                    "table": distribution_table_rows,
                    "schema": {"type": "TableType"},
                }
            )
            value_type = "TableType"

        return RenderedAtomicContent(
            name="atomic.diagnostic.observed_value",
            value=value_obj,
            value_type=value_type,
        )

    @classmethod
    @renderer(renderer_type=LegacyDiagnosticRendererType.OBSERVED_VALUE)
    def _diagnostic_observed_value_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        if not result.result.get("details"):
            return "--"

        observed_partition_object = result.result["details"]["observed_partition"]
        observed_distribution = cls._get_kl_divergence_chart(observed_partition_object)

        observed_value = (
            num_to_str(result.result.get("observed_value"))
            if result.result.get("observed_value")
            else result.result.get("observed_value")
        )

        observed_value_content_block = RenderedStringTemplateContent(
            **{
                "content_block_type": "string_template",
                "string_template": {
                    "template": "KL Divergence: $observed_value",
                    "params": {
                        "observed_value": (
                            str(observed_value)
                            if observed_value
                            else "None (-infinity, infinity, or NaN)"
                        ),
                    },
                    "styling": {"classes": ["mb-2"]},
                },
            }
        )

        return RenderedContentBlockContainer(
            **{
                "content_block_type": "content_block_container",
                "content_blocks": [observed_value_content_block, observed_distribution],
            }
        )

    @classmethod
    @renderer(renderer_type=LegacyDescriptiveRendererType.HISTOGRAM)
    def _descriptive_histogram_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        assert result, "Must pass in result."
        observed_partition_object = result.result["details"]["observed_partition"]
        weights = observed_partition_object["weights"]
        if len(weights) > 60:  # noqa: PLR2004 # FIXME CoP
            return None

        header = RenderedStringTemplateContent(
            **{
                "content_block_type": "string_template",
                "string_template": {
                    "template": "Histogram",
                    "tooltip": {"content": "expect_column_kl_divergence_to_be_less_than"},
                    "tag": "h6",
                },
            }
        )

        return cls._get_kl_divergence_chart(observed_partition_object, header)



from __future__ import annotations

import warnings
from contextlib import contextmanager
from typing import TYPE_CHECKING, List

import sqlalchemy as sa

import great_expectations as gx
from great_expectations.core.batch_definition import BatchDefinition
from great_expectations.datasource.fluent import GxDatasourceWarning
from great_expectations.execution_engine.sqlalchemy_batch_data import (
    SqlAlchemyBatchData,
)
from tests.integration.fixtures.partition_and_sample_data.partitioner_test_cases_and_fixtures import (  # noqa: E501 # FIXME CoP
    TaxiPartitioningTestCase,
    TaxiPartitioningTestCasesBase,
)
from tests.test_utils import (
    LoadedTable,
    add_datasource,
    clean_up_tables_with_prefix,
    load_and_concatenate_csvs,
    load_data_into_test_database,
)

if TYPE_CHECKING:
    import pandas as pd

TAXI_DATA_TABLE_NAME: str = "taxi_data_all_samples"


def _load_data(
    connection_string: str,
    dialect: str,
    table_name: str = TAXI_DATA_TABLE_NAME,
    random_table_suffix: bool = True,
) -> LoadedTable:
    dialects_supporting_multiple_values_in_single_insert_clause: List[str] = ["redshift"]
    to_sql_method: str = (
        "multi" if dialect in dialects_supporting_multiple_values_in_single_insert_clause else None
    )

    # Load the first 10 rows of each month of taxi data
    return load_data_into_test_database(
        table_name=table_name,
        csv_paths=[
            "./data/ten_trips_from_each_month/yellow_tripdata_sample_10_trips_from_each_month.csv"
        ],
        connection_string=connection_string,
        convert_colnames_to_datetime=["pickup_datetime", "dropoff_datetime"],
        load_full_dataset=True,
        random_table_suffix=random_table_suffix,
        to_sql_method=to_sql_method,
    )


def _is_dialect_athena(dialect: str) -> bool:
    """Is the dialect awsathena?"""
    return dialect == "awsathena"


@contextmanager
def loaded_table(dialect: str, connection_string: str) -> LoadedTable:
    test_df: pd.DataFrame
    table_name: str
    loaded_table: LoadedTable
    if _is_dialect_athena(dialect):
        table_name = "ten_trips_from_each_month"
        test_df = load_and_concatenate_csvs(
            csv_paths=[
                "./data/ten_trips_from_each_month/yellow_tripdata_sample_10_trips_from_each_month.csv"
            ],
            convert_column_names_to_datetime=["pickup_datetime", "dropoff_datetime"],
            load_full_dataset=True,
        )
        loaded_table = LoadedTable(
            table_name=table_name,
            inserted_dataframe=test_df,
        )
    else:
        loaded_table = _load_data(connection_string=connection_string, dialect=dialect)

    try:
        yield loaded_table
    finally:
        if not _is_dialect_athena(dialect):
            print("Cleaning up created loaded table")
            clean_up_tables_with_prefix(
                connection_string=connection_string,
                table_prefix=loaded_table.table_name,
            )


def _execute_taxi_partitioning_test_cases(
    taxi_partitioning_test_cases: TaxiPartitioningTestCasesBase,
    connection_string: str,
    table_name: str,
) -> None:
    test_cases: List[TaxiPartitioningTestCase] = taxi_partitioning_test_cases.test_cases()

    test_case: TaxiPartitioningTestCase
    for test_case in test_cases:
        print("Testing add_batch_definition_* method", test_case.add_batch_definition_method_name)

        # 1. Setup

        context = gx.get_context(mode="ephemeral")

        datasource_name: str = "test_datasource"
        batch_definition_name: str = "test_batch_definition"
        data_asset_name: str = table_name  # Read from generated table name

        column_name: str = taxi_partitioning_test_cases.test_column_name
        column_names: List[str] = taxi_partitioning_test_cases.test_column_names

        # 2. Set partitioner in DataConnector config
        datasource = add_datasource(
            context, name=datasource_name, connection_string=connection_string
        )

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "once",
                message="The `schema_name argument` is deprecated",
                category=DeprecationWarning,
            )
            warnings.filterwarnings(
                "once",
                message="schema_name None does not match datasource schema",
                category=GxDatasourceWarning,
            )

            asset = datasource.add_table_asset(
                data_asset_name, table_name=table_name, schema_name=None
            )

        add_batch_definition_method = getattr(
            asset, test_case.add_batch_definition_method_name or "MAKE THIS REQUIRED"
        )
        batch_definition: BatchDefinition = add_batch_definition_method(
            name=batch_definition_name, **test_case.add_batch_definition_kwargs
        )

        # 3. Check if resulting batches are as expected
        batch_request = batch_definition.build_batch_request()
        batch_identifiers_list = asset.get_batch_identifiers_list(batch_request)
        assert len(batch_identifiers_list) == test_case.num_expected_batch_definitions, (
            f"Found {len(batch_identifiers_list)} batch definitions "
            f"but expected {test_case.num_expected_batch_definitions}"
        )

        expected_batch_metadata: List[dict]

        if test_case.table_domain_test_case:
            expected_batch_metadata = [{}]
        elif column_name or column_names:
            # This condition is a smell. Consider refactoring.
            expected_batch_metadata = [data for data in test_case.expected_column_values]
        else:
            raise ValueError("Missing test_column_names or test_column_names attribute.")

        assert batch_identifiers_list == expected_batch_metadata, (
            f"Batch metadata lists don't match.\n\n"
            f"batch_list:\n{batch_identifiers_list}\n\n"
            f"expected_batch metadata:\n{expected_batch_metadata}"
        )

        # 4. Check that loaded data is as expected, using correctness
        # of arbitrary batch as a proxy for correctness of the whole list

        batch = asset.get_batch(batch_request)
        execution_engine = datasource.get_execution_engine()
        batch_data: SqlAlchemyBatchData = execution_engine.get_batch_data(
            batch_spec=batch.batch_spec
        )

        num_rows: int = execution_engine.execute_query(
            sa.select(sa.func.count()).select_from(batch_data.selectable)
        ).scalar()
        assert num_rows == test_case.num_expected_rows_in_first_batch_definition



from copy import deepcopy

import pytest

from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.exceptions import (
    DataContextError,
    InvalidExpectationConfigurationError,
)
from great_expectations.expectations.expectation_configuration import (
    ExpectationConfiguration,
)


@pytest.fixture
def empty_suite() -> ExpectationSuite:
    return ExpectationSuite(
        name="warning",
        expectations=[],
        meta={"notes": "This is an expectation suite."},
    )


@pytest.fixture
def baseline_suite(exp1, exp2) -> ExpectationSuite:
    return ExpectationSuite(
        name="warning",
        expectations=[exp1, exp2],
        meta={"notes": "This is an expectation suite."},
    )


@pytest.fixture
def exp1() -> ExpectationConfiguration:
    return ExpectationConfiguration(
        type="expect_column_values_to_be_in_set",
        kwargs={"column": "a", "value_set": [1, 2, 3], "result_format": "BASIC"},
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def exp2() -> ExpectationConfiguration:
    return ExpectationConfiguration(
        type="expect_column_values_to_be_in_set",
        kwargs={"column": "b", "value_set": [-1, -2, -3], "result_format": "BASIC"},
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def exp3() -> ExpectationConfiguration:
    return ExpectationConfiguration(
        type="expect_column_values_to_be_in_set",
        kwargs={"column": "b", "value_set": [-1, -2, -3], "result_format": "BASIC"},
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def exp4() -> ExpectationConfiguration:
    return ExpectationConfiguration(
        type="expect_column_values_to_be_in_set",
        kwargs={"column": "b", "value_set": [1, 2, 3], "result_format": "BASIC"},
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def exp5() -> ExpectationConfiguration:
    return ExpectationConfiguration(
        type="expect_column_values_to_be_in_set",
        kwargs={"column": "b", "value_set": [1, 2, 3], "result_format": "COMPLETE"},
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def exp6() -> ExpectationConfiguration:
    return ExpectationConfiguration(
        type="expect_column_values_to_be_in_set",
        kwargs={"column": "b", "value_set": [1, 2]},
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def exp7() -> ExpectationConfiguration:
    return ExpectationConfiguration(
        type="expect_column_values_to_be_in_set",
        kwargs={"column": "b", "value_set": [1, 2, 3, 4]},
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def exp8() -> ExpectationConfiguration:
    return ExpectationConfiguration(
        type="expect_column_values_to_be_in_set",
        kwargs={"column": "b", "value_set": [1, 2, 3]},
        meta={"notes": "This is an expectation."},
    )


@pytest.fixture
def table_exp1() -> ExpectationConfiguration:
    return ExpectationConfiguration(
        type="expect_table_columns_to_match_ordered_list",
        kwargs={"column_list": ["a", "b", "c"]},
    )


@pytest.fixture
def table_exp2() -> ExpectationConfiguration:
    return ExpectationConfiguration(
        type="expect_table_row_count_to_be_between",
        kwargs={"min_value": 0, "max_value": 1},
    )


@pytest.fixture
def table_exp3() -> ExpectationConfiguration:
    return ExpectationConfiguration(type="expect_table_row_count_to_equal", kwargs={"value": 1})


@pytest.fixture
def column_pair_expectation() -> ExpectationConfiguration:
    return ExpectationConfiguration(
        type="expect_column_pair_values_to_be_in_set",
        kwargs={
            "column_A": "1",
            "column_B": "b",
            "value_pairs_set": [(1, 1), (2, 2)],
            "result_format": "BASIC",
        },
    )


@pytest.fixture
def single_expectation_suite(exp1) -> ExpectationSuite:
    return ExpectationSuite(
        name="warning",
        expectations=[exp1],
        meta={"notes": "This is an expectation suite."},
    )


@pytest.fixture
def single_expectation_suite_with_expectation_ge_cloud_id(exp1) -> ExpectationSuite:
    exp1_with_ge_cloud_id = deepcopy(exp1)
    exp1_with_ge_cloud_id.id = "0faf94a9-f53a-41fb-8e94-32f218d4a774"

    return ExpectationSuite(
        name="warning",
        expectations=[exp1_with_ge_cloud_id],
        meta={"notes": "This is an expectation suite."},
    )


@pytest.fixture
def different_suite(exp1, exp4) -> ExpectationSuite:
    return ExpectationSuite(
        name="warning",
        expectations=[exp1, exp4],
        meta={"notes": "This is an expectation suite."},
    )


@pytest.fixture
def domain_success_runtime_suite(exp1, exp2, exp3, exp4, exp5) -> ExpectationSuite:
    return ExpectationSuite(
        name="warning",
        expectations=[exp1, exp2, exp3, exp4, exp5],
        meta={"notes": "This is an expectation suite."},
    )


@pytest.fixture
def suite_with_table_and_column_expectations(
    exp1,
    exp2,
    exp3,
    exp4,
    column_pair_expectation,
    table_exp1,
    table_exp2,
    table_exp3,
) -> ExpectationSuite:
    suite = ExpectationSuite(
        name="warning",
        expectations=[
            exp1,
            exp2,
            exp3,
            exp4,
            column_pair_expectation,
            table_exp1,
            table_exp2,
            table_exp3,
        ],
        meta={"notes": "This is an expectation suite."},
    )
    return suite


@pytest.fixture
def suite_with_column_pair_and_table_expectations(
    table_exp1,
    table_exp2,
    table_exp3,
    column_pair_expectation,
) -> ExpectationSuite:
    suite = ExpectationSuite(
        name="warning",
        expectations=[
            column_pair_expectation,
            table_exp1,
            table_exp2,
            table_exp3,
        ],
        meta={"notes": "This is an expectation suite."},
    )
    assert suite.expectation_configurations == [
        column_pair_expectation,
        table_exp1,
        table_exp2,
        table_exp3,
    ]
    return suite


@pytest.fixture
def ge_cloud_suite(ge_cloud_id, exp1, exp2, exp3) -> ExpectationSuite:
    for exp in (exp1, exp2, exp3):
        exp.id = ge_cloud_id
    return ExpectationSuite(
        name="warning",
        expectations=[exp1, exp2, exp3],
        meta={"notes": "This is an expectation suite."},
        id=ge_cloud_id,
    )


@pytest.mark.filesystem
def test_remove_expectation(
    exp1, exp2, exp3, exp4, exp5, single_expectation_suite, domain_success_runtime_suite
):
    domain_success_runtime_suite.remove_expectation(
        exp5, match_type="runtime", remove_multiple_matches=False
    )  # remove one matching expectation

    with pytest.raises(ValueError):
        domain_success_runtime_suite.remove_expectation(exp5, match_type="runtime")

    assert len(domain_success_runtime_suite.expectations) == 4

    with pytest.raises(ValueError):
        domain_success_runtime_suite.remove_expectation(
            exp4, match_type="domain", remove_multiple_matches=False
        )

    # remove 3 matching expectations
    domain_success_runtime_suite.remove_expectation(
        exp4, match_type="domain", remove_multiple_matches=True
    )

    with pytest.raises(ValueError):
        domain_success_runtime_suite.remove_expectation(exp2, match_type="runtime")
    with pytest.raises(ValueError):
        domain_success_runtime_suite.remove_expectation(exp3, match_type="runtime")

    assert len(domain_success_runtime_suite.expectations) == 1
    assert domain_success_runtime_suite == single_expectation_suite


@pytest.mark.filesystem
def test_remove_expectation_without_necessary_args(single_expectation_suite):
    with pytest.raises(TypeError) as err:
        single_expectation_suite.remove_expectation(expectation_configuration=None, id=None)
    assert str(err.value) == "Must provide either expectation_configuration or id"


@pytest.mark.filesystem
def test_add_expectation_configurations(
    exp1,
    exp2,
    exp3,
    exp4,
    exp5,
    single_expectation_suite,
    different_suite,
):
    expectation_configurations = [exp1, exp2, exp3, exp4, exp5]
    assert len(single_expectation_suite.expectations) == 1
    assert single_expectation_suite != different_suite
    result = single_expectation_suite.add_expectation_configurations(
        expectation_configurations=expectation_configurations,
        match_type="domain",
        overwrite_existing=True,
    )
    assert len(result) == 5

    # Collisions/overrites due to same "match_type" value
    assert len(single_expectation_suite.expectations) == 2

    # Should raise if overwrite_existing=False and a matching expectation is found
    with pytest.raises(DataContextError):
        # noinspection PyUnusedLocal
        result = single_expectation_suite.add_expectation_configurations(
            expectation_configurations=expectation_configurations,
            match_type="domain",
            overwrite_existing=False,
        )

    assert len(single_expectation_suite.expectations) == len(different_suite.expectations)
    assert all(
        mine.isEquivalentTo(theirs)
        for (mine, theirs) in zip(
            single_expectation_suite.expectation_configurations,
            different_suite.expectation_configurations,
        )
    )


@pytest.mark.filesystem
def test_add_expectation(
    exp2,
    exp4,
    single_expectation_suite,
    baseline_suite,
    different_suite,
    domain_success_runtime_suite,
):
    assert len(single_expectation_suite.expectations) == 1
    assert single_expectation_suite != baseline_suite
    single_expectation_suite.add_expectation_configuration(
        exp2, match_type="runtime", overwrite_existing=False
    )
    assert single_expectation_suite == baseline_suite
    assert len(single_expectation_suite.expectations) == 2

    # Should raise if overwrite_existing=False and a matching expectation is found
    with pytest.raises(DataContextError):
        single_expectation_suite.add_expectation_configuration(
            exp4, match_type="domain", overwrite_existing=False
        )

    assert single_expectation_suite != different_suite
    single_expectation_suite.add_expectation_configuration(
        exp4, match_type="domain", overwrite_existing=True
    )
    assert single_expectation_suite == different_suite
    assert len(single_expectation_suite.expectations) == 2

    # Should raise if more than one matching expectation is found
    with pytest.raises(ValueError):
        domain_success_runtime_suite.add_expectation_configuration(
            exp2, match_type="success", overwrite_existing=False
        )

    config = ExpectationConfiguration(type="not an expectation", kwargs={})
    with pytest.raises(InvalidExpectationConfigurationError):
        single_expectation_suite.add_expectation_configuration(config)


@pytest.mark.cloud
def test_add_expectation_with_ge_cloud_id(
    single_expectation_suite_with_expectation_ge_cloud_id,
):
    """
    This test ensures that expectation does not lose ge_cloud_id attribute when updated
    """
    expectation_ge_cloud_id = (
        single_expectation_suite_with_expectation_ge_cloud_id.expectation_configurations[0].id
    )
    # updated expectation does not have ge_cloud_id
    updated_expectation = ExpectationConfiguration(
        type="expect_column_values_to_be_in_set",
        kwargs={
            "column": "a",
            "value_set": [11, 22, 33, 44, 55],
            "result_format": "BASIC",
        },
        meta={"notes": "This is an expectation."},
    )
    single_expectation_suite_with_expectation_ge_cloud_id.add_expectation_configuration(
        updated_expectation, overwrite_existing=True
    )
    assert (
        single_expectation_suite_with_expectation_ge_cloud_id.expectation_configurations[0].id
        == expectation_ge_cloud_id
    )
    # make sure expectation config was actually updated
    assert single_expectation_suite_with_expectation_ge_cloud_id.expectation_configurations[
        0
    ].kwargs["value_set"] == [
        11,
        22,
        33,
        44,
        55,
    ]



from ..util import verify_dynamic_loading_support as _verify_dynamic_loading_support
from .actions import (
    EmailAction,
    MicrosoftTeamsNotificationAction,
    OpsgenieAlertAction,
    PagerdutyAlertAction,
    SlackNotificationAction,
    SNSNotificationAction,
    UpdateDataDocsAction,
    ValidationAction,
)
from .checkpoint import ActionContext, Checkpoint, CheckpointResult

for _module_name, _package_name in [
    (".actions", "great_expectations.checkpoint"),
    (".checkpoint", "great_expectations.checkpoint"),
]:
    _verify_dynamic_loading_support(module_name=_module_name, package_name=_package_name)

# cleanup namespace
del _verify_dynamic_loading_support
del _module_name
del _package_name



from __future__ import annotations

import logging
import os
import re
from typing import TYPE_CHECKING, Callable, ClassVar, List, Optional, Type

from great_expectations.compatibility import azure, pydantic
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.batch_spec import AzureBatchSpec, PathBatchSpec
from great_expectations.datasource.fluent.data_connector.file_path_data_connector import (
    FilePathDataConnector,
    MissingFilePathTemplateMapFnError,
)

if TYPE_CHECKING:
    from great_expectations.alias_types import PathStr
    from great_expectations.compatibility import azure
    from great_expectations.core.batch import LegacyBatchDefinition


logger = logging.getLogger(__name__)


class _AzureOptions(pydantic.BaseModel):
    abs_container: str
    abs_name_starts_with: str = ""
    abs_delimiter: str = "/"
    abs_recursive_file_discovery: bool = False


class AzureBlobStorageDataConnector(FilePathDataConnector):
    """Extension of FilePathDataConnector used to connect to Microsoft Azure Blob Storage (ABS).

    Args:
        datasource_name: The name of the Datasource associated with this DataConnector instance
        data_asset_name: The name of the DataAsset using this DataConnector instance
        azure_client: Reference to instantiated Microsoft Azure Blob Storage client handle
        account_name (str): account name for Microsoft Azure Blob Storage
        container (str): container name for Microsoft Azure Blob Storage
        name_starts_with (str): Microsoft Azure Blob Storage prefix
        delimiter (str): Microsoft Azure Blob Storage delimiter
        recursive_file_discovery (bool): Flag to indicate if files should be searched recursively from subfolders
        file_path_template_map_fn: Format function mapping path to fully-qualified resource on ABS
        whole_directory_path_override: If present, treat entire directory as single Asset
    """  # noqa: E501 # FIXME CoP

    asset_level_option_keys: ClassVar[tuple[str, ...]] = (
        "abs_container",
        "abs_name_starts_with",
        "abs_delimiter",
        "abs_recursive_file_discovery",
    )
    asset_options_type: ClassVar[Type[_AzureOptions]] = _AzureOptions

    def __init__(  # noqa: PLR0913 # FIXME CoP
        self,
        datasource_name: str,
        data_asset_name: str,
        azure_client: azure.BlobServiceClient,
        account_name: str,
        container: str,
        name_starts_with: str = "",
        delimiter: str = "/",
        recursive_file_discovery: bool = False,
        file_path_template_map_fn: Optional[Callable] = None,
        whole_directory_path_override: PathStr | None = None,
    ) -> None:
        self._azure_client: azure.BlobServiceClient = azure_client

        self._account_name = account_name
        self._container = container

        self._prefix: str = name_starts_with
        self._sanitized_prefix: str = sanitize_prefix(text=name_starts_with)

        self._delimiter = delimiter

        self._recursive_file_discovery = recursive_file_discovery

        super().__init__(
            datasource_name=datasource_name,
            data_asset_name=data_asset_name,
            file_path_template_map_fn=file_path_template_map_fn,
            whole_directory_path_override=whole_directory_path_override,
        )

    @classmethod
    def build_data_connector(  # noqa: PLR0913 # FIXME CoP
        cls,
        datasource_name: str,
        data_asset_name: str,
        azure_client: azure.BlobServiceClient,
        account_name: str,
        container: str,
        name_starts_with: str = "",
        delimiter: str = "/",
        recursive_file_discovery: bool = False,
        file_path_template_map_fn: Optional[Callable] = None,
        whole_directory_path_override: PathStr | None = None,
    ) -> AzureBlobStorageDataConnector:
        """Builds "AzureBlobStorageDataConnector", which links named DataAsset to Microsoft Azure Blob Storage.

        Args:
            datasource_name: The name of the Datasource associated with this "AzureBlobStorageDataConnector" instance
            data_asset_name: The name of the DataAsset using this "AzureBlobStorageDataConnector" instance
            azure_client: Reference to instantiated Microsoft Azure Blob Storage client handle
            account_name: account name for Microsoft Azure Blob Storage
            container: container name for Microsoft Azure Blob Storage
            name_starts_with: Microsoft Azure Blob Storage prefix
            delimiter: Microsoft Azure Blob Storage delimiter
            recursive_file_discovery: Flag to indicate if files should be searched recursively from subfolders
            file_path_template_map_fn: Format function mapping path to fully-qualified resource on ABS
            whole_directory_path_override: If present, treat entire directory as single Asset

        Returns:
            Instantiated "AzureBlobStorageDataConnector" object
        """  # noqa: E501 # FIXME CoP
        return AzureBlobStorageDataConnector(
            datasource_name=datasource_name,
            data_asset_name=data_asset_name,
            azure_client=azure_client,
            account_name=account_name,
            container=container,
            name_starts_with=name_starts_with,
            delimiter=delimiter,
            recursive_file_discovery=recursive_file_discovery,
            file_path_template_map_fn=file_path_template_map_fn,
            whole_directory_path_override=whole_directory_path_override,
        )

    @classmethod
    def build_test_connection_error_message(  # noqa: PLR0913 # FIXME CoP
        cls,
        data_asset_name: str,
        account_name: str,
        container: str,
        name_starts_with: str = "",
        delimiter: str = "/",
        recursive_file_discovery: bool = False,
    ) -> str:
        """Builds helpful error message for reporting issues when linking named DataAsset to Microsoft Azure Blob Storage.

        Args:
            data_asset_name: The name of the DataAsset using this "AzureBlobStorageDataConnector" instance
            account_name: account name for Microsoft Azure Blob Storage
            container: container name for Microsoft Azure Blob Storage
            name_starts_with: Microsoft Azure Blob Storage prefix
            delimiter: Microsoft Azure Blob Storage delimiter
            recursive_file_discovery: Flag to indicate if files should be searched recursively from subfolders

        Returns:
            Customized error message
        """  # noqa: E501 # FIXME CoP
        test_connection_error_message_template: str = 'No file belonging to account "{account_name}" in container "{container}" with prefix "{name_starts_with}" and recursive file discovery set to "{recursive_file_discovery}" found using delimiter "{delimiter}" for DataAsset "{data_asset_name}".'  # noqa: E501 # FIXME CoP
        return test_connection_error_message_template.format(
            **{
                "data_asset_name": data_asset_name,
                "account_name": account_name,
                "container": container,
                "name_starts_with": name_starts_with,
                "delimiter": delimiter,
                "recursive_file_discovery": recursive_file_discovery,
            }
        )

    @override
    def build_batch_spec(self, batch_definition: LegacyBatchDefinition) -> AzureBatchSpec:
        """
        Build BatchSpec from batch_definition by calling DataConnector's build_batch_spec function.

        Args:
            batch_definition (LegacyBatchDefinition): to be used to build batch_spec

        Returns:
            BatchSpec built from batch_definition
        """
        batch_spec: PathBatchSpec = super().build_batch_spec(batch_definition=batch_definition)
        return AzureBatchSpec(batch_spec)

    # Interface Method
    @override
    def get_data_references(self) -> List[str]:
        query_options: dict = {
            "container": self._container,
            "name_starts_with": self._sanitized_prefix,
            "delimiter": self._delimiter,
        }
        path_list: List[str] = list_azure_keys(
            azure_client=self._azure_client,
            query_options=query_options,
            recursive=self._recursive_file_discovery,
        )
        return path_list

    # Interface Method
    @override
    def _get_full_file_path(self, path: str) -> str:
        # If the path is already a fully qualified Azure URL (starts with wasbs://), return it as-is
        # This handles the case of whole_directory_path_override which is already fully qualified
        if path.startswith("wasbs://"):
            return path

        if self._file_path_template_map_fn is None:
            raise MissingFilePathTemplateMapFnError()

        template_arguments = {
            "account_name": self._account_name,
            "container": self._container,
            "path": path,
        }

        return self._file_path_template_map_fn(**template_arguments)

    @override
    def _preprocess_batching_regex(self, regex: re.Pattern) -> re.Pattern:
        regex = re.compile(f"{re.escape(self._sanitized_prefix)}{regex.pattern}")
        return super()._preprocess_batching_regex(regex=regex)


def sanitize_prefix(text: str) -> str:
    """
    Takes in a given user-prefix and cleans it to work with file-system traversal methods
    (i.e. add '/' to the end of a string meant to represent a directory)
    """
    _, ext = os.path.splitext(text)  # noqa: PTH122 # FIXME CoP
    if ext:
        # Provided prefix is a filename so no adjustment is necessary
        return text

    # Provided prefix is a directory (so we want to ensure we append it with '/')
    return os.path.join(text, "")  # noqa: PTH118 # FIXME CoP


def list_azure_keys(
    azure_client: azure.BlobServiceClient,
    query_options: dict,
    recursive: bool = False,
) -> List[str]:
    """
    Utilizes the Azure Blob Storage connection object to retrieve blob names based on user-provided criteria.

    For InferredAssetAzureDataConnector, we take container and name_starts_with and search for files using RegEx at and below the level
    specified by those parameters. However, for ConfiguredAssetAzureDataConnector, we take container and name_starts_with and
    search for files using RegEx only at the level specified by that bucket and prefix.

    This restriction for the ConfiguredAssetAzureDataConnector is needed, because paths on Azure are comprised not only the leaf file name
    but the full path that includes both the prefix and the file name.  Otherwise, in the situations where multiple data assets
    share levels of a directory tree, matching files to data assets will not be possible, due to the path ambiguity.

    Args:
        azure_client (BlobServiceClient): Azure connnection object responsible for accessing container
        query_options (dict): Azure query attributes ("container", "name_starts_with", "delimiter")
        recursive (bool): True for InferredAssetAzureDataConnector and False for ConfiguredAssetAzureDataConnector (see above)

    Returns:
        List of keys representing Azure file paths (as filtered by the query_options dict)
    """  # noqa: E501 # FIXME CoP
    container: str = query_options["container"]
    container_client: azure.ContainerClient = azure_client.get_container_client(container=container)

    path_list: List[str] = []

    def _walk_blob_hierarchy(name_starts_with: str) -> None:
        for item in container_client.walk_blobs(name_starts_with=name_starts_with):
            if isinstance(item, azure.BlobPrefix):
                if recursive:
                    _walk_blob_hierarchy(name_starts_with=item.name)

            else:
                path_list.append(item.name)

    name_starts_with: str = query_options["name_starts_with"]
    _walk_blob_hierarchy(name_starts_with)

    return path_list



from unittest.mock import ANY

import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    JUST_PANDAS_DATA_SOURCES,
    NON_SQL_DATA_SOURCES,
    SQL_DATA_SOURCES,
)

BASIC_COL = "basic"
DISTRIBUTION_WITH_OUTLIER = "with_outlier"
MOSTLY_ZERO_DISTRIBUTION = "mostly_zero"
DISTRIBUTION_WITH_NULLS = "lotta_nulls"

DATA = pd.DataFrame(
    {
        BASIC_COL: [1, 1, 1, 3, 3],
        DISTRIBUTION_WITH_OUTLIER: [-1000000, -1, 0, 1, 1],
        MOSTLY_ZERO_DISTRIBUTION: [1, 0, 0, 0, 0],
        DISTRIBUTION_WITH_NULLS: [-1, 0, 1, None, None],
    },
    dtype="object",
)


@parameterize_batch_for_data_sources(data_source_configs=NON_SQL_DATA_SOURCES, data=DATA)
def test_success_complete__non_sql(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnValueZScoresToBeLessThan(
        column=BASIC_COL, threshold=1.96, double_sided=True
    )
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert result.success


@parameterize_batch_for_data_sources(data_source_configs=SQL_DATA_SOURCES, data=DATA)
def test_success_complete__sql(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnValueZScoresToBeLessThan(
        column=BASIC_COL, threshold=1.96, double_sided=True
    )
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert result.success
    assert result.to_json_dict()["result"] == {
        "element_count": 5,
        "unexpected_count": 0,
        "unexpected_percent": 0.0,
        "partial_unexpected_list": [],
        "missing_count": 0,
        "missing_percent": 0.0,
        "unexpected_percent_total": 0.0,
        "unexpected_percent_nonmissing": 0.0,
        "partial_unexpected_counts": [],
        "unexpected_list": [],
        "unexpected_index_query": ANY,
    }


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnValueZScoresToBeLessThan(
                column=BASIC_COL, threshold=1.96, double_sided=True
            ),
            id="basic_successful_test",
        ),
        pytest.param(
            gxe.ExpectColumnValueZScoresToBeLessThan(
                column=BASIC_COL, threshold=1.96, double_sided=True, mostly=0.8
            ),
            id="successful_test_with_mostly",
        ),
        pytest.param(
            gxe.ExpectColumnValueZScoresToBeLessThan(
                column=DISTRIBUTION_WITH_OUTLIER, threshold=1, double_sided=True, mostly=0.6
            ),
            id="outlier_test_with_mostly",
        ),
        pytest.param(
            gxe.ExpectColumnValueZScoresToBeLessThan(
                column=BASIC_COL, threshold=1.96, double_sided=False
            ),
            id="single_sided_test",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success(
    batch_for_datasource: Batch,
    expectation: gxe.ExpectColumnValueZScoresToBeLessThan,
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnValueZScoresToBeLessThan(
                column=BASIC_COL, threshold=0.1, double_sided=True
            ),
            id="threshold_too_low",
        ),
        pytest.param(
            gxe.ExpectColumnValueZScoresToBeLessThan(
                column=DISTRIBUTION_WITH_OUTLIER, threshold=1, double_sided=True
            ),
            id="extreme_outlier",
        ),
        pytest.param(
            gxe.ExpectColumnValueZScoresToBeLessThan(
                column=MOSTLY_ZERO_DISTRIBUTION, threshold=1, double_sided=True, mostly=0.9
            ),
            id="mostly_requirement_not_met",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_failure(
    batch_for_datasource: Batch,
    expectation: gxe.ExpectColumnValueZScoresToBeLessThan,
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert not result.success
